###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

DCM_Field_Lookup = {
    'Video_Unmutes':
        'INTEGER',
    'Zip_Postal_Code':
        'STRING',
    'Path_Length':
        'INTEGER',
    'Measurable_Impressions_For_Audio':
        'INTEGER',
    'Average_Interaction_Time':
        'FLOAT',
    'Invalid_Impressions':
        'INTEGER',
    'Floodlight_Variable_40':
        'STRING',
    'Floodlight_Variable_41':
        'STRING',
    'Floodlight_Variable_42':
        'STRING',
    'Floodlight_Variable_43':
        'STRING',
    'Floodlight_Variable_44':
        'STRING',
    'Floodlight_Variable_45':
        'STRING',
    'Floodlight_Variable_46':
        'STRING',
    'Floodlight_Variable_47':
        'STRING',
    'Floodlight_Variable_48':
        'STRING',
    'Floodlight_Variable_49':
        'STRING',
    'Clicks':
        'INTEGER',
    'Warnings':
        'INTEGER',
    'Paid_Search_Advertiser_Id':
        'INTEGER',
    'Video_Third_Quartile_Completions':
        'INTEGER',
    'Video_Progress_Events':
        'INTEGER',
    'Cost_Per_Revenue':
        'FLOAT',
    'Total_Interaction_Time':
        'INTEGER',
    'Roadblock_Impressions':
        'INTEGER',
    'Video_Replays':
        'INTEGER',
    'Keyword':
        'STRING',
    'Full_Screen_Video_Plays':
        'INTEGER',
    'Active_View_Impression_Distribution_Not_Measurable':
        'FLOAT',
    'Unique_Reach_Total_Reach':
        'INTEGER',
    'Has_Exits':
        'BOOLEAN',
    'Paid_Search_Engine_Account':
        'STRING',
    'Operating_System_Version':
        'STRING',
    'Campaign':
        'STRING',
    'Active_View_Percent_Visible_At_Start':
        'FLOAT',
    'Active_View_Not_Viewable_Impressions':
        'INTEGER',
    'Twitter_Offers_Accepted':
        'INTEGER',
    'Interaction_Type':
        'STRING',
    'Activity_Delivery_Status':
        'FLOAT',
    'Video_Companion_Clicks':
        'INTEGER',
    'Floodlight_Paid_Search_Average_Cost_Per_Transaction':
        'FLOAT',
    'Paid_Search_Agency_Id':
        'INTEGER',
    'Asset':
        'STRING',
    'Active_View_Audible_Amp_Fully_On_Screen_For_Half_Of_Duration_15_Sec_Cap_Impressions':
        'INTEGER',
    'User_List_Current_Size':
        'STRING',
    'Exit_Url':
        'STRING',
    'Natural_Search_Revenue':
        'FLOAT',
    'Retweets':
        'INTEGER',
    'Full_Screen_Impressions':
        'INTEGER',
    'Audio_Unmutes':
        'INTEGER',
    'Dynamic_Element_4_Field_Value_2':
        'STRING',
    'Dynamic_Element_4_Field_Value_3':
        'STRING',
    'Dynamic_Element_4_Field_Value_1':
        'STRING',
    'Creative_Start_Date':
        'DATE',
    'Small_Video_Player_Size_Impressions':
        'INTEGER',
    'Audio_Replays':
        'INTEGER',
    'Video_Player_Size_Avg_Width':
        'INTEGER',
    'Rich_Media_Standard_Event_Count':
        'INTEGER',
    'Publisher_Problems':
        'INTEGER',
    'Paid_Search_Advertiser':
        'STRING',
    'Has_Video_Completions':
        'BOOLEAN',
    'Cookie_Reach_Duplicate_Impression_Reach':
        'INTEGER',
    'Audio_Third_Quartile_Completions':
        'INTEGER',
    'Package_Roadblock_Strategy':
        'STRING',
    'Video_Midpoints':
        'INTEGER',
    'Click_Through_Conversion_Events_Cross_Environment':
        'INTEGER',
    'Video_Pauses':
        'INTEGER',
    'Trueview_Views':
        'INTEGER',
    'Interaction_Count_Mobile_Static_Image':
        'INTEGER',
    'Dynamic_Element_Click_Rate':
        'FLOAT',
    'Dynamic_Element_1_Field_6_Value':
        'STRING',
    'Dynamic_Element_4_Value':
        'STRING',
    'Creative_End_Date':
        'DATE',
    'Dynamic_Element_4_Field_3_Value':
        'STRING',
    'Dynamic_Field_Value_3':
        'STRING',
    'Mobile_Carrier':
        'STRING',
    'U_Value':
        'STRING',
    'Average_Display_Time':
        'FLOAT',
    'Custom_Variable':
        'STRING',
    'Video_Interactions':
        'INTEGER',
    'Average_Expansion_Time':
        'FLOAT',
    'Email_Shares':
        'INTEGER',
    'Flight_Booked_Rate':
        'STRING',
    'Impression_Count':
        'INTEGER',
    'Site_Id_Dcm':
        'INTEGER',
    'Dynamic_Element_3_Value_Id':
        'STRING',
    'Paid_Search_Legacy_Keyword_Id':
        'INTEGER',
    'View_Through_Conversion_Events_Cross_Environment':
        'INTEGER',
    'Counters':
        'INTEGER',
    'Floodlight_Paid_Search_Average_Cost_Per_Action':
        'FLOAT',
    'Activity_Group_Id':
        'INTEGER',
    'Active_View_Percent_Audible_And_Visible_At_Midpoint':
        'FLOAT',
    'Click_Count':
        'INTEGER',
    'Video_4A_39_S_Ad_Id':
        'STRING',
    'Total_Interactions':
        'INTEGER',
    'Active_View_Viewable_Impressions':
        'INTEGER',
    'Natural_Search_Engine_Property':
        'STRING',
    'Video_Views':
        'INTEGER',
    'Domain':
        'STRING',
    'Total_Conversion_Events_Cross_Environment':
        'INTEGER',
    'Dbm_Advertiser':
        'STRING',
    'Companion_Impressions':
        'INTEGER',
    'View_Through_Conversions_Cross_Environment':
        'INTEGER',
    'Dynamic_Element_5_Field_Value_3':
        'STRING',
    'Dynamic_Element_5_Field_Value_2':
        'STRING',
    'Dynamic_Element_5_Field_Value_1':
        'STRING',
    'Dynamic_Field_Value_1':
        'STRING',
    'Active_View_Impression_Distribution_Viewable':
        'FLOAT',
    'Creative_Field_2':
        'STRING',
    'Twitter_Leads_Generated':
        'INTEGER',
    'Paid_Search_External_Campaign_Id':
        'INTEGER',
    'Creative_Field_8':
        'STRING',
    'Interaction_Count_Paid_Search':
        'INTEGER',
    'Activity_Group':
        'STRING',
    'Video_Player_Location_Avg_Pixels_From_Top':
        'INTEGER',
    'Interaction_Number':
        'INTEGER',
    'Dynamic_Element_3_Value':
        'STRING',
    'Cookie_Reach_Total_Reach':
        'INTEGER',
    'Cookie_Reach_Exclusive_Click_Reach':
        'INTEGER',
    'Creative_Field_5':
        'STRING',
    'Cookie_Reach_Overlap_Impression_Reach':
        'INTEGER',
    'Paid_Search_Ad_Group':
        'STRING',
    'Interaction_Count_Rich_Media':
        'INTEGER',
    'Dynamic_Element_Total_Conversions':
        'INTEGER',
    'Transaction_Count':
        'INTEGER',
    'Num_Value':
        'STRING',
    'Cookie_Reach_Overlap_Total_Reach':
        'INTEGER',
    'Floodlight_Attribution_Type':
        'STRING',
    'Cookie_Reach_Exclusive_Impression_Reach_Percent':
        'FLOAT',
    'Html5_Impressions':
        'INTEGER',
    'Served_Pixel_Density':
        'STRING',
    'Has_Full_Screen_Video_Plays':
        'BOOLEAN',
    'Interaction_Count_Static_Image':
        'INTEGER',
    'Has_Video_Companion_Clicks':
        'BOOLEAN',
    'Twitter_Video_100Percent_In_View_For_3_Seconds':
        'INTEGER',
    'Paid_Search_Ad':
        'STRING',
    'Paid_Search_Visits':
        'INTEGER',
    'Audio_Pauses':
        'INTEGER',
    'Creative_Pixel_Size':
        'STRING',
    'Flight_Start_Date':
        'DATE',
    'Active_View_Percent_Of_Third_Quartile_Impressions_Visible':
        'FLOAT',
    'Natural_Search_Transactions':
        'FLOAT',
    'Cookie_Reach_Impression_Reach':
        'INTEGER',
    'Dynamic_Field_Value_4':
        'STRING',
    'Twitter_Line_Item_Id':
        'INTEGER',
    'Has_Video_Stops':
        'BOOLEAN',
    'Active_View_Percent_Audible_And_Visible_At_Completion':
        'FLOAT',
    'Paid_Search_Labels':
        'STRING',
    'Twitter_Creative_Type':
        'STRING',
    'Operating_System':
        'STRING',
    'Dynamic_Element_3_Field_4_Value':
        'STRING',
    'Hours_Since_First_Interaction':
        'INTEGER',
    'Asset_Category':
        'STRING',
    'Active_View_Measurable_Impressions':
        'INTEGER',
    'Package_Roadblock':
        'STRING',
    'Large_Video_Player_Size_Impressions':
        'INTEGER',
    'Paid_Search_Actions':
        'FLOAT',
    'Active_View_Percent_Visible_At_Midpoint':
        'FLOAT',
    'Has_Full_Screen_Views':
        'BOOLEAN',
    'Backup_Image':
        'INTEGER',
    'Likes':
        'INTEGER',
    'Dbm_Line_Item':
        'STRING',
    'Active_View_Percent_Audible_Impressions':
        'FLOAT',
    'Flight_Booked_Cost':
        'STRING',
    'Other_Twitter_Engagements':
        'INTEGER',
    'Audio_Completions':
        'INTEGER',
    'Click_Rate':
        'FLOAT',
    'Cost_Per_Activity':
        'FLOAT',
    'Dynamic_Element_2_Value':
        'STRING',
    'Cookie_Reach_Exclusive_Impression_Reach':
        'INTEGER',
    'Content_Category':
        'STRING',
    'Twitter_Video_50Percent_In_View_For_2_Seconds':
        'INTEGER',
    'Total_Revenue_Cross_Environment':
        'FLOAT',
    'Unique_Reach_Click_Reach':
        'INTEGER',
    'Has_Video_Replays':
        'BOOLEAN',
    'Twitter_Creative_Id':
        'INTEGER',
    'Has_Counters':
        'BOOLEAN',
    'Dynamic_Element_4_Value_Id':
        'STRING',
    'Dynamic_Element_4_Field_1_Value':
        'STRING',
    'Has_Video_Interactions':
        'BOOLEAN',
    'Video_Player_Size':
        'STRING',
    'Active_View_Audible_Amp_Fully_On_Screen_For_Half_Of_Duration_Trueview_Impressions':
        'INTEGER',
    'Advertiser_Id':
        'INTEGER',
    'Rich_Media_Clicks':
        'INTEGER',
    'Floodlight_Variable_59':
        'STRING',
    'Floodlight_Variable_58':
        'STRING',
    'Floodlight_Variable_57':
        'STRING',
    'Floodlight_Variable_56':
        'STRING',
    'Floodlight_Variable_55':
        'STRING',
    'Floodlight_Variable_54':
        'STRING',
    'Floodlight_Variable_53':
        'STRING',
    'Floodlight_Variable_52':
        'STRING',
    'Floodlight_Variable_51':
        'STRING',
    'Floodlight_Variable_50':
        'STRING',
    'Cookie_Reach_Incremental_Total_Reach':
        'INTEGER',
    'Playback_Method':
        'STRING',
    'Has_Interactive_Impressions':
        'BOOLEAN',
    'Paid_Search_Average_Position':
        'FLOAT',
    'Floodlight_Variable_96':
        'STRING',
    'Package_Roadblock_Id':
        'INTEGER',
    'Recalculated_Attribution_Type':
        'STRING',
    'Dbm_Advertiser_Id':
        'INTEGER',
    'Floodlight_Variable_100':
        'STRING',
    'Cookie_Reach_Exclusive_Total_Reach_Percent':
        'FLOAT',
    'Active_View_Impression_Distribution_Not_Viewable':
        'FLOAT',
    'Floodlight_Variable_3':
        'STRING',
    'Floodlight_Variable_2':
        'STRING',
    'Floodlight_Variable_1':
        'STRING',
    'Floodlight_Variable_7':
        'STRING',
    'Floodlight_Variable_6':
        'STRING',
    'Floodlight_Variable_5':
        'STRING',
    'Floodlight_Variable_4':
        'STRING',
    'Rendering_Id':
        'INTEGER',
    'Floodlight_Variable_9':
        'STRING',
    'Floodlight_Variable_8':
        'STRING',
    'Cookie_Reach_Incremental_Impression_Reach':
        'INTEGER',
    'Active_View_Percent_Visible_At_Third_Quartile':
        'FLOAT',
    'Reporting_Problems':
        'INTEGER',
    'Package_Roadblock_Total_Booked_Units':
        'STRING',
    'Active_View_Percent_Visible_At_Completion':
        'FLOAT',
    'Has_Video_Midpoints':
        'BOOLEAN',
    'Dynamic_Element_4_Field_4_Value':
        'STRING',
    'Percentage_Of_Measurable_Impressions_For_Video_Player_Location':
        'FLOAT',
    'Campaign_End_Date':
        'DATE',
    'Placement_External_Id':
        'STRING',
    'Cost_Per_Click':
        'FLOAT',
    'Cookie_Reach_Overlap_Click_Reach_Percent':
        'FLOAT',
    'Active_View_Percent_Full_Screen':
        'FLOAT',
    'Hour':
        'STRING',
    'Click_Through_Revenue':
        'FLOAT',
    'Video_Skips':
        'INTEGER',
    'Paid_Search_Click_Rate':
        'FLOAT',
    'Has_Video_Views':
        'BOOLEAN',
    'Dbm_Cost_Account_Currency':
        'FLOAT',
    'Flight_End_Date':
        'DATE',
    'Has_Video_Plays':
        'BOOLEAN',
    'Paid_Search_Clicks':
        'INTEGER',
    'Creative_Field_9':
        'STRING',
    'Manual_Closes':
        'INTEGER',
    'Creative_Field_7':
        'STRING',
    'Creative_Field_6':
        'STRING',
    'App':
        'STRING',
    'Creative_Field_4':
        'STRING',
    'Creative_Field_3':
        'STRING',
    'Campaign_Start_Date':
        'DATE',
    'Creative_Field_1':
        'STRING',
    'Content_Classifier':
        'STRING',
    'Cookie_Reach_Duplicate_Click_Reach':
        'INTEGER',
    'Site_Dcm':
        'STRING',
    'Digital_Content_Label':
        'STRING',
    'Has_Manual_Closes':
        'BOOLEAN',
    'Has_Timers':
        'BOOLEAN',
    'Impressions':
        'INTEGER',
    'Classified_Impressions':
        'INTEGER',
    'Dbm_Site_Id':
        'INTEGER',
    'Dynamic_Element_2_Field_1_Value':
        'STRING',
    'Floodlight_Variable_72':
        'STRING',
    'Creative':
        'STRING',
    'Asset_Orientation':
        'STRING',
    'Custom_Variable_Count_1':
        'INTEGER',
    'Video_Stops':
        'INTEGER',
    'Paid_Search_Ad_Id':
        'INTEGER',
    'Cookie_Reach_Overlap_Total_Reach_Percent':
        'FLOAT',
    'Click_Delivery_Status':
        'FLOAT',
    'Dynamic_Element_Impressions':
        'INTEGER',
    'Interaction_Count_Click_Tracker':
        'INTEGER',
    'Placement_Total_Booked_Units':
        'STRING',
    'Date':
        'DATE',
    'Twitter_Placement_Type':
        'STRING',
    'Total_Revenue':
        'FLOAT',
    'Recalculated_Attributed_Interaction':
        'STRING',
    'Ad_Type':
        'STRING',
    'Social_Engagement_Rate':
        'FLOAT',
    'Dynamic_Element_5_Field_1_Value':
        'STRING',
    'Dynamic_Profile_Id':
        'INTEGER',
    'Active_View_Impressions_Visible_10_Seconds':
        'INTEGER',
    'Interaction_Count_Video':
        'INTEGER',
    'Dynamic_Element_5_Value':
        'STRING',
    'Video_Player_Size_Avg_Height':
        'INTEGER',
    'Creative_Type':
        'STRING',
    'Campaign_External_Id':
        'STRING',
    'Dynamic_Element_Click_Through_Conversions':
        'INTEGER',
    'Conversion_Url':
        'STRING',
    'Floodlight_Variable_89':
        'STRING',
    'Floodlight_Variable_84':
        'STRING',
    'Floodlight_Variable_85':
        'STRING',
    'Floodlight_Variable_86':
        'STRING',
    'Floodlight_Variable_87':
        'STRING',
    'Floodlight_Variable_80':
        'STRING',
    'Floodlight_Variable_81':
        'STRING',
    'Floodlight_Variable_82':
        'STRING',
    'Floodlight_Variable_83':
        'STRING',
    'Active_View_Audible_Amp_Fully_On_Screen_For_Half_Of_Duration_Trueview_Measurable_Impressions':
        'INTEGER',
    'Floodlight_Variable_66':
        'STRING',
    'Floodlight_Variable_67':
        'STRING',
    'Floodlight_Variable_64':
        'STRING',
    'Floodlight_Variable_65':
        'STRING',
    'Floodlight_Variable_62':
        'STRING',
    'Floodlight_Variable_63':
        'STRING',
    'Floodlight_Variable_60':
        'STRING',
    'Active_View_Eligible_Impressions':
        'INTEGER',
    'Dynamic_Element_3_Field_Value_1':
        'STRING',
    'Dynamic_Element_3_Field_Value_3':
        'STRING',
    'Dynamic_Element_3_Field_Value_2':
        'STRING',
    'Floodlight_Variable_68':
        'STRING',
    'Floodlight_Variable_69':
        'STRING',
    'Floodlight_Variable_13':
        'STRING',
    'Floodlight_Variable_12':
        'STRING',
    'Floodlight_Variable_11':
        'STRING',
    'Floodlight_Variable_10':
        'STRING',
    'Floodlight_Variable_17':
        'STRING',
    'Floodlight_Variable_16':
        'STRING',
    'Floodlight_Variable_15':
        'STRING',
    'Floodlight_Variable_14':
        'STRING',
    'Floodlight_Variable_19':
        'STRING',
    'Floodlight_Variable_18':
        'STRING',
    'Audience_Targeted':
        'STRING',
    'Total_Conversions_Cross_Environment':
        'INTEGER',
    'Interaction_Count_Mobile_Rich_Media':
        'INTEGER',
    'Active_View_Percent_Viewable_Impressions':
        'FLOAT',
    'Dynamic_Element_1_Field_1_Value':
        'STRING',
    'Active_View_Audible_Amp_Fully_On_Screen_For_Half_Of_Duration_Trueview_Rate':
        'FLOAT',
    'Has_Video_Skips':
        'BOOLEAN',
    'Dynamic_Element_5_Field_2_Value':
        'STRING',
    'Twitter_Buy_Now_Clicks':
        'INTEGER',
    'Creative_Groups_2':
        'STRING',
    'Creative_Groups_1':
        'STRING',
    'Campaign_Id':
        'INTEGER',
    'Twitter_Campaign_Id':
        'INTEGER',
    'Active_View_Audible_Amp_Fully_On_Screen_For_Half_Of_Duration_15_Sec_Cap_Rate':
        'FLOAT',
    'Dynamic_Element_1_Field_5_Value':
        'STRING',
    'Paid_Search_Match_Type':
        'STRING',
    'Activity_Per_Thousand_Impressions':
        'FLOAT',
    'Has_Expansions':
        'BOOLEAN',
    'Dbm_Creative_Id':
        'INTEGER',
    'Active_View_Percent_Audible_And_Visible_At_Start':
        'FLOAT',
    'Booked_Viewable_Impressions':
        'FLOAT',
    'Dynamic_Element_1_Field_2_Value':
        'STRING',
    'Paid_Search_Cost':
        'FLOAT',
    'Dynamic_Element_5_Field_5_Value':
        'STRING',
    'Floodlight_Paid_Search_Spend_Per_Transaction_Revenue':
        'FLOAT',
    'Active_View_Percent_Of_Midpoint_Impressions_Audible_And_Visible':
        'FLOAT',
    'Dynamic_Element_4':
        'STRING',
    'Dynamic_Element_5':
        'STRING',
    'Dynamic_Element_2':
        'STRING',
    'Click_Through_Conversions':
        'FLOAT',
    'Dynamic_Element_1':
        'STRING',
    'Dynamic_Element_4_Field_6_Value':
        'STRING',
    'Attributed_Event_Platform_Type':
        'STRING',
    'Audio_Midpoints':
        'INTEGER',
    'Attributed_Event_Connection_Type':
        'STRING',
    'Dynamic_Element_1_Value':
        'STRING',
    'Measurable_Impressions_For_Video_Player_Location':
        'INTEGER',
    'Audio_Companion_Impressions':
        'INTEGER',
    'Video_Full_Screen':
        'INTEGER',
    'Active_View_Percent_Visible_At_First_Quartile':
        'FLOAT',
    'Companion_Creative':
        'STRING',
    'Cookie_Reach_Exclusive_Total_Reach':
        'INTEGER',
    'Audio_Mutes':
        'INTEGER',
    'Placement_Rate':
        'STRING',
    'Companion_Clicks':
        'INTEGER',
    'Cookie_Reach_Overlap_Click_Reach':
        'INTEGER',
    'Site_Keyname':
        'STRING',
    'Placement_Cost_Structure':
        'STRING',
    'Percentage_Of_Measurable_Impressions_For_Audio':
        'FLOAT',
    'Rich_Media_Custom_Event_Count':
        'INTEGER',
    'Dbm_Cost_Usd':
        'FLOAT',
    'Dynamic_Element_1_Field_3_Value':
        'STRING',
    'Has_Video_Third_Quartile_Completions':
        'BOOLEAN',
    'Paid_Search_Landing_Page_Url':
        'STRING',
    'Verifiable_Impressions':
        'INTEGER',
    'Average_Time':
        'FLOAT',
    'Creative_Field_12':
        'STRING',
    'Creative_Field_11':
        'STRING',
    'Creative_Field_10':
        'STRING',
    'Channel_Mix':
        'STRING',
    'Paid_Search_Campaign':
        'STRING',
    'Natural_Search_Landing_Page':
        'STRING',
    'Dynamic_Element_1_Field_4_Value':
        'STRING',
    'Payment_Source':
        'STRING',
    'Planned_Media_Cost':
        'FLOAT',
    'Conversion_Referrer':
        'STRING',
    'Companion_Creative_Id':
        'INTEGER',
    'Dynamic_Element_4_Field_2_Value':
        'STRING',
    'Total_Conversions':
        'FLOAT',
    'Custom_Variable_Count_2':
        'INTEGER',
    'Paid_Search_External_Ad_Group_Id':
        'INTEGER',
    'Hd_Video_Player_Size_Impressions':
        'INTEGER',
    'Click_Through_Transaction_Count':
        'FLOAT',
    'Cookie_Reach_Exclusive_Click_Reach_Percent':
        'FLOAT',
    'Floodlight_Attributed_Interaction':
        'STRING',
    'Dynamic_Profile':
        'STRING',
    'Floodlight_Variable_28':
        'STRING',
    'Floodlight_Variable_29':
        'STRING',
    'Dynamic_Element_2_Field_2_Value':
        'STRING',
    'Floodlight_Variable_22':
        'STRING',
    'Floodlight_Variable_23':
        'STRING',
    'Floodlight_Variable_20':
        'STRING',
    'Floodlight_Variable_21':
        'STRING',
    'Floodlight_Variable_26':
        'STRING',
    'Floodlight_Variable_27':
        'STRING',
    'Floodlight_Variable_24':
        'STRING',
    'Floodlight_Variable_25':
        'STRING',
    'Dynamic_Element_4_Field_5_Value':
        'STRING',
    'Creative_Id':
        'INTEGER',
    'Activity_Per_Click':
        'FLOAT',
    'Floodlight_Variable_88':
        'STRING',
    'Active_View_Percent_Of_First_Quartile_Impressions_Visible':
        'FLOAT',
    'Serving_Problems':
        'INTEGER',
    'Placement':
        'STRING',
    'Dynamic_Element_2_Field_Value_1':
        'STRING',
    'Dynamic_Element_2_Field_Value_2':
        'STRING',
    'Dynamic_Element_2_Field_Value_3':
        'STRING',
    'Interaction_Count_Mobile_Video':
        'INTEGER',
    'Has_Full_Screen_Video_Completions':
        'BOOLEAN',
    'Placement_Total_Planned_Media_Cost':
        'STRING',
    'Video_First_Quartile_Completions':
        'INTEGER',
    'Twitter_Creative_Media_Id':
        'INTEGER',
    'Cookie_Reach_Duplicate_Total_Reach':
        'INTEGER',
    'Rich_Media_Impressions':
        'INTEGER',
    'Video_Completions':
        'INTEGER',
    'Month':
        'STRING',
    'Paid_Search_Keyword_Id':
        'INTEGER',
    'Cookie_Reach_Duplicate_Click_Reach_Percent':
        'FLOAT',
    'Replies':
        'INTEGER',
    'Dynamic_Element_5_Field_6_Value':
        'STRING',
    'Video_Mutes':
        'INTEGER',
    'Flight_Booked_Units':
        'STRING',
    'Dynamic_Element_Value_Id':
        'STRING',
    'Expansion_Time':
        'INTEGER',
    'Invalid_Clicks':
        'INTEGER',
    'Has_Video_Progress_Events':
        'BOOLEAN',
    'Dynamic_Element_2_Field_3_Value':
        'STRING',
    'Rich_Media_Standard_Event_Path_Summary':
        'STRING',
    'Video_Muted_At_Start':
        'INTEGER',
    'Audio_Companion_Clicks':
        'INTEGER',
    'Active_View_Audible_Amp_Fully_On_Screen_For_Half_Of_Duration_15_Sec_Cap_Measurable_Impressions':
        'INTEGER',
    'Interaction_Date_Time':
        'DATETIME',
    'User_List':
        'STRING',
    'Event_Timers':
        'FLOAT',
    'Paid_Search_External_Keyword_Id':
        'INTEGER',
    'Timers':
        'INTEGER',
    'Floodlight_Paid_Search_Action_Conversion_Percentage':
        'FLOAT',
    'View_Through_Revenue_Cross_Environment':
        'FLOAT',
    'Advertiser':
        'STRING',
    'Has_Video_Unmutes':
        'BOOLEAN',
    'Natural_Search_Query':
        'STRING',
    'Audio_Plays':
        'INTEGER',
    'Unique_Reach_Average_Impression_Frequency':
        'FLOAT',
    'Path_Type':
        'STRING',
    'Dynamic_Field_Value_2':
        'STRING',
    'Interaction_Channel':
        'STRING',
    'Blocked_Impressions':
        'INTEGER',
    'Dynamic_Field_Value_5':
        'STRING',
    'Dynamic_Field_Value_6':
        'STRING',
    'Placement_Compatibility':
        'STRING',
    'City':
        'STRING',
    'Dbm_Line_Item_Id':
        'INTEGER',
    'Cookie_Reach_Incremental_Click_Reach':
        'INTEGER',
    'Floodlight_Variable_61':
        'STRING',
    'Natural_Search_Processed_Landing_Page_Query_String':
        'STRING',
    'Report_Day':
        'DATE',
    'Dbm_Site':
        'STRING',
    'Connection_Type':
        'STRING',
    'Video_Average_View_Time':
        'FLOAT',
    'Click_Through_Revenue_Cross_Environment':
        'FLOAT',
    'Dbm_Creative':
        'STRING',
    'Attributed_Event_Environment':
        'STRING',
    'Floodlight_Variable_99':
        'STRING',
    'Floodlight_Variable_98':
        'STRING',
    'Measurable_Impressions_For_Video_Player_Size':
        'INTEGER',
    'Dynamic_Element_1_Value_Id':
        'STRING',
    'Paid_Search_Engine_Account_Id':
        'INTEGER',
    'Floodlight_Variable_93':
        'STRING',
    'Floodlight_Variable_92':
        'STRING',
    'Floodlight_Variable_91':
        'STRING',
    'Floodlight_Variable_90':
        'STRING',
    'Floodlight_Variable_97':
        'STRING',
    'Placement_Strategy':
        'STRING',
    'Floodlight_Variable_95':
        'STRING',
    'Floodlight_Variable_94':
        'STRING',
    'Active_View_Percent_Of_Completed_Impressions_Audible_And_Visible':
        'FLOAT',
    'Floodlight_Variable_75':
        'STRING',
    'Floodlight_Variable_74':
        'STRING',
    'Floodlight_Variable_77':
        'STRING',
    'Floodlight_Variable_76':
        'STRING',
    'Floodlight_Variable_71':
        'STRING',
    'Floodlight_Variable_70':
        'STRING',
    'Floodlight_Variable_73':
        'STRING',
    'Activity':
        'STRING',
    'Natural_Search_Engine_Url':
        'STRING',
    'Total_Display_Time':
        'INTEGER',
    'User_List_Description':
        'STRING',
    'Active_View_Impressions_Audible_And_Visible_At_Completion':
        'INTEGER',
    'Floodlight_Variable_79':
        'STRING',
    'Floodlight_Variable_78':
        'STRING',
    'Twitter_Impression_Type':
        'STRING',
    'Active_View_Average_Viewable_Time_Seconds':
        'FLOAT',
    'Active_View_Percent_Measurable_Impressions':
        'FLOAT',
    'Natural_Search_Landing_Page_Query_String':
        'STRING',
    'Percentage_Of_Measurable_Impressions_For_Video_Player_Size':
        'FLOAT',
    'Has_Full_Screen_Impressions':
        'BOOLEAN',
    'Conversion_Id':
        'INTEGER',
    'Creative_Version':
        'STRING',
    'Active_View_Percent_In_Background':
        'FLOAT',
    'Dynamic_Element_2_Value_Id':
        'STRING',
    'Hours_Since_Attributed_Interaction':
        'INTEGER',
    'Dynamic_Element_3_Field_5_Value':
        'STRING',
    'Has_Video_First_Quartile_Completions':
        'BOOLEAN',
    'Dynamic_Element':
        'STRING',
    'Active_View_Percent_Visible_10_Seconds':
        'FLOAT',
    'Booked_Clicks':
        'FLOAT',
    'Booked_Impressions':
        'FLOAT',
    'Tran_Value':
        'STRING',
    'Dynamic_Element_Clicks':
        'INTEGER',
    'Has_Dynamic_Impressions':
        'BOOLEAN',
    'Site_Id_Site_Directory':
        'INTEGER',
    'Active_View_Percent_Audible_And_Visible_At_Third_Quartile':
        'FLOAT',
    'Twitter_Url_Clicks':
        'INTEGER',
    'Dynamic_Element_2_Field_6_Value':
        'STRING',
    'Has_Backup_Image':
        'BOOLEAN',
    'Dynamic_Element_2_Field_5_Value':
        'STRING',
    'Rich_Media_Custom_Event_Path_Summary':
        'STRING',
    'Advertiser_Group':
        'STRING',
    'General_Invalid_Traffic_Givt_Clicks':
        'INTEGER',
    'Has_Video_Pauses':
        'BOOLEAN',
    'Follows':
        'INTEGER',
    'Has_Html5_Impressions':
        'BOOLEAN',
    'Percent_Invalid_Clicks':
        'FLOAT',
    'Percent_Invalid_Impressions':
        'FLOAT',
    'Activity_Date_Time':
        'DATETIME',
    'Site_Site_Directory':
        'STRING',
    'Placement_Pixel_Size':
        'STRING',
    'Within_Floodlight_Lookback_Window':
        'STRING',
    'Dbm_Partner':
        'STRING',
    'Dynamic_Element_3_Field_3_Value':
        'STRING',
    'Ord_Value':
        'STRING',
    'Floodlight_Configuration':
        'INTEGER',
    'Ad_Id':
        'INTEGER',
    'Video_Plays':
        'INTEGER',
    'Video_Player_Location_Avg_Pixels_From_Left':
        'INTEGER',
    'Days_Since_First_Interaction':
        'INTEGER',
    'Event_Counters':
        'INTEGER',
    'Active_View_Not_Measurable_Impressions':
        'INTEGER',
    'Landing_Page_Url':
        'STRING',
    'Ad_Status':
        'STRING',
    'Unique_Reach_Impression_Reach':
        'INTEGER',
    'Dynamic_Element_2_Field_4_Value':
        'STRING',
    'Dbm_Partner_Id':
        'INTEGER',
    'Asset_Id':
        'INTEGER',
    'Video_View_Rate':
        'FLOAT',
    'Twitter_App_Install_Clicks':
        'INTEGER',
    'Cookie_Reach_Duplicate_Total_Reach_Percent':
        'FLOAT',
    'Total_Social_Engagements':
        'INTEGER',
    'Media_Cost':
        'FLOAT',
    'Placement_Tag_Type':
        'STRING',
    'Dbm_Insertion_Order':
        'STRING',
    'Floodlight_Variable_39':
        'STRING',
    'Floodlight_Variable_38':
        'STRING',
    'Paid_Search_External_Ad_Id':
        'INTEGER',
    'Browser_Platform':
        'STRING',
    'Floodlight_Variable_31':
        'STRING',
    'Floodlight_Variable_30':
        'STRING',
    'Floodlight_Variable_33':
        'STRING',
    'Floodlight_Variable_32':
        'STRING',
    'Floodlight_Variable_35':
        'STRING',
    'Floodlight_Variable_34':
        'STRING',
    'Floodlight_Variable_37':
        'STRING',
    'Floodlight_Variable_36':
        'STRING',
    'Dynamic_Element_View_Through_Conversions':
        'INTEGER',
    'Active_View_Viewable_Impression_Cookie_Reach':
        'INTEGER',
    'Video_Interaction_Rate':
        'FLOAT',
    'Dynamic_Element_3_Field_1_Value':
        'STRING',
    'Booked_Activities':
        'FLOAT',
    'Has_Video_Full_Screen':
        'BOOLEAN',
    'User_List_Membership_Life_Span':
        'STRING',
    'Video_Length':
        'STRING',
    'Paid_Search_Keyword':
        'STRING',
    'Revenue_Per_Click':
        'FLOAT',
    'Downloaded_Impressions':
        'INTEGER',
    'Days_Since_Attributed_Interaction':
        'INTEGER',
    'Code_Serves':
        'INTEGER',
    'Effective_Cpm':
        'FLOAT',
    'Environment':
        'STRING',
    'Active_View_Percent_Play_Time_Audible_And_Visible':
        'FLOAT',
    'Paid_Search_Agency':
        'STRING',
    'Dynamic_Element_5_Field_3_Value':
        'STRING',
    'Paid_Search_Engine_Account_Category':
        'STRING',
    'Week':
        'STRING',
    'Designated_Market_Area_Dma':
        'STRING',
    'Cookie_Reach_Click_Reach':
        'INTEGER',
    'Twitter_Buy_Now_Purchases':
        'INTEGER',
    'Floodlight_Paid_Search_Average_Dcm_Transaction_Amount':
        'FLOAT',
    'Placement_Start_Date':
        'DATE',
    'Cookie_Reach_Overlap_Impression_Reach_Percent':
        'FLOAT',
    'View_Through_Revenue':
        'FLOAT',
    'Impression_Delivery_Status':
        'FLOAT',
    'Floodlight_Paid_Search_Transaction_Conversion_Percentage':
        'FLOAT',
    'Click_Through_Conversions_Cross_Environment':
        'INTEGER',
    'Activity_Id':
        'INTEGER',
    'Has_Video_Mutes':
        'BOOLEAN',
    'Exits':
        'INTEGER',
    'Paid_Search_Bid_Strategy':
        'STRING',
    'Active_View_Percent_Of_Midpoint_Impressions_Visible':
        'FLOAT',
    'Dbm_Insertion_Order_Id':
        'INTEGER',
    'Placement_Id':
        'INTEGER',
    'App_Id':
        'STRING',
    'View_Through_Transaction_Count':
        'FLOAT',
    'Floodlight_Paid_Search_Transaction_Revenue_Per_Spend':
        'FLOAT',
    'View_Through_Conversions':
        'FLOAT',
    'Dynamic_Element_5_Field_4_Value':
        'STRING',
    'Active_View_Percent_Of_First_Quartile_Impressions_Audible_And_Visible':
        'FLOAT',
    'Companion_Creative_Pixel_Size':
        'STRING',
    'Revenue_Per_Thousand_Impressions':
        'FLOAT',
    'Natural_Search_Clicks':
        'INTEGER',
    'Dynamic_Element_3_Field_2_Value':
        'STRING',
    'Audio_First_Quartile_Completions':
        'INTEGER',
    'Dynamic_Element_1_Field_Value_3':
        'STRING',
    'Dynamic_Element_1_Field_Value_2':
        'STRING',
    'Dynamic_Element_1_Field_Value_1':
        'STRING',
    'Full_Screen_Average_View_Time':
        'FLOAT',
    'Dynamic_Element_5_Value_Id':
        'STRING',
    'Rich_Media_Click_Rate':
        'FLOAT',
    'User_List_Id':
        'INTEGER',
    'Click_Through_Url':
        'STRING',
    'Active_View_Percent_Of_Completed_Impressions_Visible':
        'FLOAT',
    'Video_Prominence_Score':
        'STRING',
    'Active_View_Percent_Play_Time_Audible':
        'FLOAT',
    'Natural_Search_Actions':
        'FLOAT',
    'Platform_Type':
        'STRING',
    'General_Invalid_Traffic_Givt_Impressions':
        'INTEGER',
    'Active_View_Percent_Audible_And_Visible_At_First_Quartile':
        'FLOAT',
    'Dynamic_Element_3':
        'STRING',
    'Active_View_Percent_Play_Time_Visible':
        'FLOAT',
    'Active_View_Percent_Of_Third_Quartile_Impressions_Audible_And_Visible':
        'FLOAT',
    'Paid_Search_Transactions':
        'FLOAT',
    'Rich_Media_Event':
        'STRING',
    'Country':
        'STRING',
    'Dynamic_Element_3_Field_6_Value':
        'STRING',
    'Expansions':
        'INTEGER',
    'Interaction_Rate':
        'FLOAT',
    'Natural_Search_Processed_Landing_Page':
        'STRING',
    'Floodlight_Impressions':
        'INTEGER',
    'Paid_Search_Bid_Strategy_Id':
        'INTEGER',
    'Interactive_Impressions':
        'INTEGER',
    'Interaction_Count_Natural_Search':
        'INTEGER',
    'Twitter_Impression_Id':
        'INTEGER',
    'Ad':
        'STRING',
    'Paid_Search_Ad_Group_Id':
        'INTEGER',
    'Paid_Search_Campaign_Id':
        'INTEGER',
    'Full_Screen_Video_Completions':
        'INTEGER',
    'Dynamic_Element_Value':
        'STRING',
    'State_Region':
        'STRING',
    'Placement_End_Date':
        'DATE',
    'Cookie_Reach_Duplicate_Impression_Reach_Percent':
        'FLOAT',
    'Paid_Search_Impressions':
        'INTEGER',
    'Cookie_Reach_Average_Impression_Frequency':
        'FLOAT',
    'Natural_Search_Engine_Country':
        'STRING',
    'Paid_Search_Revenue':
        'FLOAT',
    'Percent_Composition_Impressions':
        'FLOAT',
}



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

LineItem_Write_Schema = [{
    'name': 'Line Item Id',
    'type': 'INTEGER',
    'mode': 'REQUIRED'
}, {
    'name': 'Partner Name',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Partner Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Advertiser Name',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'IO Name',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Line Item Name',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Line Item Timestamp',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Line Item Status',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'IO Start Date',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'IO End Date',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'IO Budget Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'IO Budget Amount',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'IO Pacing',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'IO Pacing Rate',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'IO Pacing Amount',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Line Item Start Date',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Line Item End Date',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Line Item Budget Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Line Item Budget Amount',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Line Item Pacing',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Line Item Pacing Rate',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Line Item Pacing Amount',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Line Item Frequency Enabled',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Line Item Frequency Exposures',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Line Item Frequency Period',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Line Item Frequency Amount',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Bid Price',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Partner Revenue Model',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Partner Revenue Amount',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Current Audience Targeting Ids',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Current Audience Targeting Names',
    'type': 'STRING',
    'mode': 'NULLABLE'
}]

LineItem_Read_Schema = [{
    'name': s['name'].replace(' ', '_'),
    'type': s['type'],
    'mode': s['mode'],
} for s in LineItem_Write_Schema]



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################


import hashlib
import json


class Colab:

  def __init__(self, name, version='4.0'):
    self.markdown_lines = []
    self.code_lines = []
    self.id = 0
    self.colab = {
        'license': 'Licensed under the Apache License, Version 2.0',
        'copyright': 'Copyright 2020 Google LLC',
        'nbformat': int(version.split('.', 1)[0]),
        'nbformat_minor': int(version.split('.', 1)[1]),
        'metadata': {
            'colab': {
                'name': name,
                'provenance': [],
                'collapsed_sections': [],
                'toc_visible': True
            },
            'kernelspec': {
                'name': 'python3',
                'display_name': 'Python 3'
            }
        },
        'cells': []
    }

  def _next_id(self):
   self.id += 1
   return "{}-{:03d}".format(hashlib.md5(self.colab['metadata']['colab']['name'].encode('utf-8')).hexdigest()[-8:], self.id)

  def _code(self):
    if self.code_lines:
      self.colab['cells'].append({
          'cell_type': 'code',
          'metadata': {
              "id": self._next_id(),
              'colab_type': 'code'
          },
          'source': self.code_lines
      })
      self.code_lines = []

  def _markdown(self):
    if self.markdown_lines:
      self.colab['cells'].append({
          'cell_type': 'markdown',
          'metadata': {
              "id": self._next_id(),
              'colab_type': 'text'
          },
          'source': self.markdown_lines
      })
      self.markdown_lines = []

  def code(self, code):
    self._markdown()
    self.code_lines.extend(['%s\n' % c for c in code.split('\n')])

  def header(self, text, level=1, indent=0):
    self._code()
    self.markdown_lines.append('%s%s%s\n' % ('>' * indent, '#' * level, text))

  def paragraph(self, text, indent=0):
    self._code()
    self.markdown_lines.extend(
        ['%s%s\n' % ('>' * indent, t) for t in text.split('\n')])

  def image(self, name, link):
    self._code()
    self.markdown_lines.append('![%s](%s)\n' % (name, link))

  def list(self, items, ordered=True, indent=0):
    self._code()
    self.markdown_lines.extend([
        '%s %s %s\n' % ('>' * indent, '1.' if ordered else '*', t)
        for t in items
    ])

  def render(self):
    self._code()
    self._markdown()
    return json.dumps(self.colab, indent=2) + '\n'



import re
import json
import datetime
from urllib.request import urlopen

from starthinker.util.project import project
from starthinker.util.sheets import sheets_read, sheets_write, sheets_clear

URL_SUFFIX = 'cm_mmc=DIS-_-AQ-_-DIS-_-%ecid!&cm_mmca1=Core&cm_mmca2=uf&cm_mmca20=OLAMC--%ebuy!--%esid!--%epid!'
liquidus_url = 'http://api.cofactordigital.com/retail/480058ae784e2461/listings.json?limit=500&storeid=%s&promotioncode=%s&tagid=5913,5914&returnmode=full&require=ListingWithLink'
stores_url = 'http://api.cofactordigital.com/retail/480058ae784e2461/stores.json?limit=%d&offset=%d'

states = {
    'AL': 'Alabama',
    'AK': 'Alaska',
    'AZ': 'Arizona',
    'AR': 'Arkansas',
    'CA': 'California',
    'CO': 'Colorado',
    'CT': 'Connecticut',
    'DE': 'Delaware',
    'FL': 'Florida',
    'GA': 'Georgia',
    'HI': 'Hawaii',
    'ID': 'Idaho',
    'IL': 'Illinois',
    'IN': 'Indiana',
    'IA': 'Iowa',
    'KS': 'Kansas',
    'KY': 'Kentucky',
    'LA': 'Louisiana',
    'ME': 'Maine',
    'MD': 'Maryland',
    'MA': 'Massachusetts',
    'MI': 'Michigan',
    'MN': 'Minnesota',
    'MS': 'Mississippi',
    'MO': 'Missouri',
    'MT': 'Montana',
    'NE': 'Nebraska',
    'NV': 'Nevada',
    'NH': 'New Hampshire',
    'NJ': 'New Jersey',
    'NM': 'New Mexico',
    'NY': 'New York',
    'NC': 'North Carolina',
    'ND': 'North Dakota',
    'OH': 'Ohio',
    'OK': 'Oklahoma',
    'OR': 'Oregon',
    'PA': 'Pennsylvania',
    'RI': 'Rhode Island',
    'SC': 'South Carolina',
    'SD': 'South Dakota',
    'TN': 'Tennessee',
    'TX': 'Texas',
    'UT': 'Utah',
    'VT': 'Vermont',
    'VA': 'Virginia',
    'WA': 'Washington',
    'WV': 'West Virginia',
    'WI': 'Wisconsin',
    'WY': 'Wyoming',
    'AS': 'American Samoa',
    'DC': 'District of Columbia',
    'FM': 'Federated States of Micronesia',
    'GU': 'Guam',
    'MH': 'Marshall Islands',
    'MP': 'Northern Mariana Islands',
    'PW': 'Palau',
    'PR': 'Puerto Rico',
    'VI': 'Virgin Islands',
    'AE': 'Armed Forces Africa',
    'AA': 'Armed Forces Americas',
    'AE': 'Armed Forces Canada',
    'AE': 'Armed Forces Europe',
    'AE': 'Armed Forces Middle East',
    'AP': 'Armed Forces Pacific'
}


def get_offers(store_id, promotion_code, retry=5):
  try:
    url = liquidus_url % (store_id, promotion_code)
    response = urlopen(url, timeout=10)
    offers = json.loads(response.read())
    return offers
  except Exception as e:
    print('An error occurred calling Liquidus API %s' % str(e))
    if retry > 0:
      return get_offers(store_id, promotion_code, retry - 1)


def clean(store):
  zip_replaces = {'98503': '98506', '98509': '98503', '60418': '60803'}

  if store['zip'].find('-') >= 0:
    store['zip'] = store['zip'][:store['zip'].find('-')]

  store['zip'] = zip_replaces.get(store['zip'], store['zip'])

  return store


def get_stores(retry=5):
  result = []
  limit = 1000
  offset = 0
  done = False

  try:
    while not done:
      url = stores_url % (limit, offset)
      offset += limit

      response = json.loads(urlopen(url).read())
      stores = response.get('results', [])

      for store in stores:
        result.append(
            clean({
                'id': store['id'],
                'state': store['address']['state'],
                'zip': store['address']['postalCode']
            }))

      if len(stores) < 1000:
        done = True

    return result
  except Exception as e:
    print(str(e))
    if retry > 0:
      return get_stores(retry - 1)


@project.from_parameters
def liquidus():
  # Read store ids from stores list
  # Legacy: reading stores from sheet, can be removed
  #sheet_range = project.task['stores_list']['range']
  #store_data = sheets_read(project.task['auth'],
  #project.task['stores_list']['sheet_id'],
  #sheet_range.split('!')[0],
  #sheet_range.split('!')[1])
  #stores = [{'id': str(item[0]), 'zip': item[6], 'state': item[5]} for item in store_data]

  stores = get_stores()

  # Read category tree
  cat_tree_range = project.task['category_tree']['range']
  raw_category_tree = sheets_read(
      project.task['auth'],
      cat_tree_range.split('!')[0],
      cat_tree_range.split('!')[1],
      sheets_url=project.task['category_tree']['sheet_id'])

  category_tree = {}
  for category in raw_category_tree:
    if len(category) >= 4:
      category_tree[int(category[0])] = {
          'CategoryTreePathForwards': category[1],
          'audience': category[2],
          'dcm_targeting_Key': category[3]
      }

  # Determine promotion code for the current week
  today = datetime.date.today()
  offset = (today.weekday() + 1) % 7
  previous_sunday = today - datetime.timedelta(offset)
  promotion_code = previous_sunday.strftime('officedepot-%y%m%d')

  # Call liquidus API for each store and add to the feed
  feed = []
  row = 0
  store_ct = 0

  # Use this in case you want to just process the first store, useful for
  # testing while developing
  # for store in stores[:1]:
  for store in stores:

    store_ct += 1

    while len(store['zip']) < 5:
      store['zip'] = '0' + store['zip']

    offers = get_offers(store['id'], promotion_code)

    print('processing store %s' % store['id'])

    for offer in offers['results']:
      feed.append([])

      startDate = datetime.datetime.fromtimestamp(
          int(re.search('\(([\d]*)\)', offer['saleStartDate']).group(1)) / 1000)
      endDate = datetime.datetime.fromtimestamp(
          int(re.search('\(([\d]*)\)', offer['saleEndDate']).group(1)) / 1000)

      days_to_save = abs(endDate - datetime.datetime.now()).days
      days_to_save_message = 'Last chance to save'
      if days_to_save == 1:
        days_to_save_message = '1 day to save'
      elif days_to_save > 1:
        days_to_save_message = '%d days to save' % days_to_save

      zip = '%s,%s,United States' % (store['zip'], states[store['state']])

      category = None

      if offer.get('departments', None):
        category = category_tree.get(offer['departments'][0]['id'], None)

      # Map API response to feed
      # TODO: Using the following line causes an overflow problem because the offer
      # ids and store ids are too long. We can't use strings because Studio
      # doesn't like it. We need to find out a way to hash these values into
      # shorter ints while keeping them unique
      # feed[row].append(int(str(offer['id']) + str(store['id'])))
      feed[row].append(int(str(offer['id']) + str(store_ct)))
      feed[row].append(offer['title'])
      feed[row].append(category['audience'] if category else '')
      feed[row].append(str(startDate))
      feed[row].append(str(endDate))
      feed[row].append(offer['deal'])
      feed[row].append(offer['originalDeal'])
      feed[row].append(offer['brands'][0]['name'])
      feed[row].append(offer['images'][0]['imageURL'])
      feed[row].append(offer['additionalDealInformation'])
      feed[row].append(zip)

      if category:
        feed[row].append(category['CategoryTreePathForwards'])
      else:
        feed[row].append('')

      link = ''

      if 'links' in offer and len(offer['links']) > 0:
        link = offer['links'][0]['linkURL']

        if not '?' in link:
          link += '?'
        else:
          link += '&'

        link += URL_SUFFIX

      feed[row].append(link)

      feed[row].append(days_to_save_message)

      if category:
        feed[row].append(category['dcm_targeting_Key'])
      else:
        feed[row].append('')

      feed[row].append(offer.get('buyOnlineLinkURL') or link)

      row += 1

  # Write feed to trix
  print('uploading feed, size %d rows' % len(feed))
  feed_range = project.task['feed']['range']
  sheets_clear(project.task['auth'], project.task['feed']['sheet_id'],
               feed_range.split('!')[0],
               feed_range.split('!')[1])
  sheets_write(project.task['auth'], project.task['feed']['sheet_id'],
               feed_range.split('!')[0],
               feed_range.split('!')[1], feed)


if __name__ == '__main__':
  liquidus()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import json
import re
from urllib import request

from googleapiclient.schema import Schemas

DATETIME_RE = re.compile(r'\d{4}[-/]\d{2}[-/]\d{2}[ T]\d{2}:\d{2}:\d{2}\.?\d+Z')
DESCRIPTION_LENGTH = 1024
RECURSION_DEPTH = 5


class Discovery_To_BigQuery():

  def __init__(self, api, version, key=None, recursion_depth=RECURSION_DEPTH):
    self.key = key or ''
    self.recursion_depth = recursion_depth
    api_url = 'https://%s.googleapis.com/$discovery/rest?version=%s&key=%s' % (
        api, version, self.key)
    print('DISCOVERY FETCH:', api_url)
    self.api_document = json.load(request.urlopen(api_url))

  @staticmethod
  def preferred_version(api_name, key=None):
    api_url = 'https://discovery.googleapis.com/discovery/v1/apis?name=%s&key=%s&preferred=true' % (
        api_name, key or '')
    print(api_url)
    api_info = json.load(request.urlopen(api_url))
    return api_info['items'][0]['version']

  @staticmethod
  def clean(struct):
    if isinstance(struct, dict):
      for key, value in struct.items():
        if isinstance(value, str) and DATETIME_RE.match(value):
          struct[key] = struct[key].replace('.000Z', 'Z')
        else:
          Discovery_To_BigQuery.clean(value)
    elif isinstance(struct, list):
      for index, value in enumerate(struct):
        if isinstance(value, str) and DATETIME_RE.match(value):
          struct[index] = struct[index].replace('.000Z', 'Z')
        else:
          Discovery_To_BigQuery.clean(value)
    return struct

  def to_type(self, entry):

    # https://developers.google.com/discovery/v1/type-format
    # https://cloud.google.com/bigquery/docs/reference/standard-sql/data-types

    t = entry.get('type')
    f = entry.get('format')

    if t == 'any':
      return 'STRING'
    elif t == 'array':
      return 'REPEATED'
    elif t == 'boolean':
      return 'BOOLEAN'
    elif t == 'integer':
      return 'INT64'
    elif t == 'number':
      if f == 'double':
        return 'NUMBER'
      else:
        return 'FLOAT'
    elif t == 'object':
      return 'STRUCT'
    elif t == 'string':
      if f == 'byte':
        return 'BYTES'
      elif f == 'date':
        return 'DATE'
      elif f == 'date-time':
        return 'DATETIME'
      elif f == 'int64':
        return 'INT64'
      elif f == 'uint64':
        return 'INT64'
      else:
        return 'STRING'
    else:
      return 'STRING'


  def to_schema(self, entry, parents={}):
    bigquery_schema = []

    for key, value in entry.items():

      if '$ref' in value:
        parents.setdefault(value['$ref'], 0)
        if parents[value['$ref']] < self.recursion_depth:
          parents[value['$ref']] += 1
          bigquery_schema.append({
              'name':
                  key,
              'type':
                  'RECORD',
              'mode':
                  'NULLABLE',
              'fields':
                  self.to_schema(
                      self.api_document['schemas'][value['$ref']]['properties'],
                      parents)
          })
        parents[value['$ref']] -= 1

      else:

        if value['type'] == 'array':

          if '$ref' in value['items']:
            parents.setdefault(value['items']['$ref'], 0)
            if parents[value['items']['$ref']] < self.recursion_depth:
              parents[value['items']['$ref']] += 1
              bigquery_schema.append({
                  'name':
                      key,
                  'type':
                      'RECORD',
                  'mode':
                      'REPEATED',
                  'fields':
                      self.to_schema(
                          self.api_document['schemas'][value['items']['$ref']]
                          ['properties'], parents)
              })
              parents[value['items']['$ref']] -= 1

          else:
            bigquery_schema.append({
                'description': (', '.join(value['items'].get('enum', [])))
                               [:DESCRIPTION_LENGTH],
                'name':
                    key,
                'type':
                    self.to_type(value['items']),
                'mode':
                    'REPEATED',
            })

        else:
          bigquery_schema.append({
              'description': (', '.join(value.get('enum', [])))
                             [:DESCRIPTION_LENGTH],
              'name':
                  key,
              'type':
                  self.to_type(value),
              'mode':
                  'NULLABLE'
          })

    return bigquery_schema


  def resource_json(self, resource):
    return self.api_document['schemas'][resource]['properties']

  def resource_schema(self, resource):
    entry = self.api_document['schemas'][resource]['properties']
    return self.to_schema(entry)


  def method_schema(self, method):
    endpoint, method = method.rsplit('.', 1)
    resource = self.api_document

    for e in endpoint.split('.'):
      resource = resource['resources'][e]
    resource = resource['methods'][method]['response']['$ref']

    # get schema
    properties = self.api_document['schemas'][resource]['properties']
    schema = self.to_schema(properties)

    # List responses wrap their items in a paginated response object
    # Unroll it to return item schema instead of repsonse schema
    if 'List' in resource and resource.endswith('Response'):
      for entry in schema:
        if entry['type'] == 'RECORD':
          return entry['fields']
        elif entry['mode'] == 'REPEATED':
          entry['mode'] = 'NULLABLE'
          return [entry]
      # raise exception after checking all fields for a list
      raise ('Unahandled discovery schema.')
    else:
      return schema


#x = Discovery_To_BigQuery('displayvideo', 'v1').resource_schema('Advertiser')
#x = Discovery_To_BigQuery('dfareporting', 'v3.4').method_schema('sites', 'list')
#print(json.dumps(x, indent=2))



# Generated by Django 2.2.13 on 2020-12-02 15:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('project', '0002_project_key'),
    ]

    operations = [
        migrations.AlterField(
            model_name='project',
            name='service',
            field=models.TextField(default=''),
        ),
    ]



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Reads weather information from the National Weather Service API

   (api.weather.gov) and outputs to a datastore such as BigQuery or sheets

"""

import json, datetime
from urllib.request import urlopen
from starthinker.util.project import project
from starthinker.util.data import put_rows

# Used to turn wind direction in degrees to compass directions
WIND_DIRECTION_MAP = {
    'NNE': {
        'f': 11.25,
        't': 33.75
    },
    'NE': {
        'f': 33.75,
        't': 56.25
    },
    'ENE': {
        'f': 56.25,
        't': 78.75
    },
    'E': {
        'f': 78.75,
        't': 101.25
    },
    'ESE': {
        'f': 101.25,
        't': 123.75
    },
    'SE': {
        'f': 123.75,
        't': 146.25
    },
    'SSE': {
        'f': 146.25,
        't': 168.75
    },
    'S': {
        'f': 168.75,
        't': 191.25
    },
    'SSW': {
        'f': 191.25,
        't': 213.75
    },
    'SW': {
        'f': 213.75,
        't': 236.25
    },
    'WSW': {
        'f': 236.25,
        't': 258.75
    },
    'W': {
        'f': 258.75,
        't': 281.25
    },
    'WNW': {
        'f': 281.25,
        't': 303.75
    },
    'NW': {
        'f': 303.75,
        't': 326.25
    },
    'NNW': {
        'f': 326.25,
        't': 348.75
    }
}

# Endpoint for last observation url of a given station
LATEST_OBSERVATION_URL = 'https://api.weather.gov/stations/%s/observations/latest'


def degrees_to_compass(value):
  """Turns direction from degrees value to compass direction

  Args:
    value: floating point representing the degrees from 0 to 360
  Returns: String representing the compass direction.
  """
  if value is None:
    return None

  if (value >= 348.75 and value <= 360) or (value >= 0 and value <= 11.25):
    return 'N'
  else:
    for direction in WIND_DIRECTION_MAP.keys():
      if value >= WIND_DIRECTION_MAP[direction][
          'f'] and value <= WIND_DIRECTION_MAP[direction]['t']:
        return direction

  return None


def c_to_f(temperature):
  """Converts temperature from celcius to fahrenheit

  Args:
    temperature: floating point representing the temperature in celcius
  Returns: temperature in fahrenheit
  """
  if temperature is None:
    return None

  return (temperature * 9 / 5) + 32


def ms_to_mph(value):
  """Converts speed from meters per second to miles per hour

  Args:
    value: floating point representing the speed in meters per second
  Returns: speed in miles per hour
  """
  if value is None:
    return None

  return value * 2.237


def m_to_mile(value):
  """Converts distance in meters to miles

  Args:
    value: floating point representing the distance in meters
  Returns: distance in miles
  """
  if value is None:
    return None

  return value / 1609


def m_to_inches(value):
  """Converts distance in meters to inches

  Args:
    value: floating point representing the distance in meters
  Returns: distance in inches
  """
  if value is None:
    return None

  return value / 39.37


@project.from_parameters
def weather_gov():
  """Main StarThinker entrypoint for weather_gov task.

  "weather_gov": {
    "auth": "user",
    "stations": ["station1", "station2"],
    "out":{
      "sheets":{
        "sheet": "sheet URL",
        "tab": "Weather",
        "range": "A2:K",
        "delete": true
      }
    }
  }

  stations: list of strings representing the station ID. This is the NOAA
  station ID, the list can be found in the weather.gov API. For example, all
  major airports are NOAA observation stations, and their station ID typically
  is K followed by the unique airport ID. E.g. The Chicago O'Hare international
  airport observation stations ID is KORD, for San Francisco international
  airport it is KSFO, Miami KMIA, etc.

  out can be any output configuration supported by StarThinker, here we have a
  sheets output as an example.

  """
  if project.verbose:
    print('weather_gov')

  rows = []

  stations = project.task.get('stations', [])

  for station_id in stations:
    result = json.loads(urlopen(LATEST_OBSERVATION_URL % station_id).read())

    rows += [[
        datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S'), station_id,
        c_to_f(result['properties']['temperature']['value']),
        c_to_f(result['properties']['windChill']['value']),
        c_to_f(result['properties']['heatIndex']['value']),
        degrees_to_compass(result['properties']['windDirection']['value']),
        ms_to_mph(result['properties']['windSpeed']['value']),
        m_to_mile(result['properties']['visibility']['value']),
        result['properties']['relativeHumidity']['value'],
        m_to_inches(result['properties']['precipitationLast3Hours']['value']),
        'NOAA'
    ]]

  put_rows(project.task['auth'], project.task['out'], rows)


if __name__ == '__main__':
  weather_gov()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.bigquery import query_to_view
from starthinker.util.bigquery import table_create
from starthinker.util.csv import rows_pad
from starthinker.util.data import get_rows
from starthinker.util.data import put_rows
from starthinker.util.google_api.discovery_to_bigquery import Discovery_To_BigQuery
from starthinker.util.project import project
from starthinker.util.regexp import lookup_id
from starthinker.util.sheets import sheets_clear

from starthinker.task.dv_editor.insertion_order import insertion_order_commit
from starthinker.task.dv_editor.patch import patch_masks
from starthinker.task.dv_editor.patch import patch_preview


def segment_clear():
  sheets_clear(
    project.task["auth_sheets"],
    project.task["sheet"],
    "Segments",
    "A2:Z"
  )


def segment_load():

  # write segments to sheet
  rows = get_rows(
    project.task["auth_bigquery"],
    { "bigquery": {
      "dataset": project.task["dataset"],
      "query": """SELECT
        CONCAT(P.displayName, ' - ', P.partnerId),
        CONCAT(A.displayName, ' - ', A.advertiserId),
        CONCAT(C.displayName, ' - ', C.campaignId),
        CONCAT(I.displayName, ' - ', I.insertionOrderId),
        'PATCH',
        CONCAT(BS.dateRange.startDate.year, '-', BS.dateRange.startDate.month, '-', BS.dateRange.startDate.day),
        CONCAT(BS.dateRange.startDate.year, '-', BS.dateRange.startDate.month, '-', BS.dateRange.startDate.day),
        CONCAT(BS.dateRange.endDate.year, '-', BS.dateRange.endDate.month, '-', BS.dateRange.endDate.day),
        CONCAT(BS.dateRange.endDate.year, '-', BS.dateRange.endDate.month, '-', BS.dateRange.endDate.day),
        BS.budgetAmountMicros / 100000,
        BS.budgetAmountMicros / 100000,
        BS.description,
        BS.description
        FROM `{dataset}.DV_InsertionOrders` AS I, UNNEST( budget.budgetSegments) AS BS
        LEFT JOIN `{dataset}.DV_Campaigns` AS C
        ON I.campaignId=C.campaignId
        LEFT JOIN `{dataset}.DV_Advertisers` AS A
        ON I.advertiserId=A.advertiserId
        LEFT JOIN `{dataset}.DV_Partners` AS P
        ON A.partnerId=P.partnerId
      """.format(**project.task),
      "legacy":False
    }}
  )

  put_rows(
    project.task["auth_sheets"],
    { "sheets": {
      "sheet": project.task["sheet"],
      "tab": "Segments",
      "range": "A2"
    }},
    rows
  )


def segment_audit():

  # Move Segments To BigQuery
  rows = get_rows(
    project.task["auth_sheets"],
    { "sheets": {
      "sheet": project.task["sheet"],
      "tab": "Segments",
      "range": "A2:M"
    }}
  )

  put_rows(
    project.task["auth_bigquery"],
    { "bigquery": {
      "dataset": project.task["dataset"],
      "table": "SHEET_Segments",
      "schema": [
        { "name": "Partner", "type": "STRING" },
        { "name": "Advertiser", "type": "STRING" },
        { "name": "Campaign", "type": "STRING" },
        { "name": "Insertion_Order", "type": "STRING" },
        { "name": "Action", "type": "STRING" },
        { "name": "Start_Date", "type": "STRING" },
        { "name": "Start_Date_Edit", "type": "STRING" },
        { "name": "End_Date", "type": "STRING" },
        { "name": "End_Date_Edit", "type": "STRING" },
        { "name": "Budget", "type": "FLOAT" },
        { "name": "Budget_Edit", "type": "FLOAT" },
        { "name": "Description", "type": "STRING" },
        { "name": "Description_Edit", "type": "STRING" },
      ],
      "format": "CSV"
    }},
    rows
  )

  # Create Audit View And Write To Sheets
  query_to_view(
    project.task["auth_bigquery"],
    project.id,
    project.task["dataset"],
    "AUDIT_Segments",
    """WITH
      /* Check if sheet values are set */
      INPUT_ERRORS AS (
        SELECT
        *
        FROM (
          SELECT
            'Segment' AS Operation,
            CASE
              WHEN Start_Date_Edit IS NULL THEN 'Missing Start Date.'
              WHEN End_Date_Edit IS NULL THEN 'Missing End Date .'
              WHEN Budget_Edit IS NULL THEN 'Missing Budget.'
            ELSE
              NULL
            END AS Error,
            'ERROR' AS Severity,
          COALESCE(Insertion_Order, 'BLANK') AS Id
        FROM
          `{dataset}.SHEET_Segments`
        )
        WHERE
          Error IS NOT NULL
      ),

      /* Check if duplicate Segments */
      DUPLICATE_ERRORS AS (
        SELECT
          'Segment' AS Operation,
          'Duplicate Segment.' AS Error,
          'WARNING' AS Severity,
          COALESCE(Insertion_Order, 'BLANK') AS Id
        FROM (
          SELECT Insertion_Order, Start_Date_Edit, End_Date_Edit, COUNT(*) AS count
          FROM `{dataset}.SHEET_Segments`
          GROUP BY 1, 2, 3
          HAVING count > 1
        )
      ),

      /* Check if budget segments are current */
      SEGMENT_ERRORS AS (
        SELECT
          *
        FROM (
          SELECT
            'Segments' AS Operation,
            CASE
              WHEN SAFE_CAST(Budget_Edit AS FLOAT64) > 1000 THEN 'Segment has excessive spend.'
              WHEN SAFE_CAST(Start_Date AS DATE) < CURRENT_DATE() THEN 'Segment starts in past.'
              WHEN SAFE_CAST(End_Date AS DATE) < CURRENT_DATE() THEN 'Segment ends in past.'
            ELSE
            NULL
          END
            AS Error,
            'ERROR' AS Severity,
            COALESCE(Insertion_Order, 'BLANK') AS Id
          FROM
            `{dataset}.SHEET_Segments`
          )
        WHERE
          Error IS NOT NULL
      )

      SELECT * FROM INPUT_ERRORS
      UNION ALL
      SELECT * FROM DUPLICATE_ERRORS
      UNION ALL
      SELECT * FROM SEGMENT_ERRORS
      ;
    """.format(**project.task),
    legacy=False
  )


def segment_patch(commit=False):

  def date_edited(value):
    y, m, d = value.split("-")
    return {"year": y, "month": m, "day": d}

  patches = {}
  changed = set()

  rows = get_rows(
    project.task["auth_sheets"],
    { "sheets": {
      "sheet": project.task["sheet"],
      "tab": "Segments",
      "range": "A2:Z"
    }}
  )

  rows = rows_pad(rows, 13, "")

  # Build list of segements skipping only deletes, track changes
  for row in rows:

    # inserts do not have an ID, skip them
    if not lookup_id(row[3]): continue

    patches.setdefault(
      row[3],
      { "operation": "Segments",
        "action": "PATCH",
        "partner": row[0],
        "advertiser": row[1],
        "campaign": row[2],
        "insertion_order": row[3],
        "parameters": {
          "advertiserId": lookup_id(row[1]),
          "insertionOrderId": lookup_id(row[3]),
          "body": {
            "budget": {
              "budgetSegments": []
            }
          }
        }
      }
    )

    if row[4] == "DELETE":
      changed.add(row[3])
    else:
      patches[row[3]]["parameters"]["body"]["budget"]["budgetSegments"].append({
        "dateRange": {
          "startDate": date_edited(row[6]),
          "endDate": date_edited(row[8])
        },
        "budgetAmountMicros": float(row[10]) * 100000,
        "description": row[12]
      })

      if row[5] != row[6] or row[7] != row[8] or row[9] != row[10] or row[11] != row[12]:
        changed.add(row[3])

  # Remove any patches where segments have not changed
  for io in list(patches.keys()):
    if io not in changed:
      del patches[io]
  patches = list(patches.values())

  patch_masks(patches)

  if commit:
    insertion_order_commit(patches)
  else:
    patch_preview(patches)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

Placement_Cost_Schema = [{
    'name': 'Advertiser_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Comments',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_End_Date',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Rate',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Start_Date',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Total_Booked_Units',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Config_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Account_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}]



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import json

from django.core.management.base import BaseCommand, CommandError

from starthinker_ui.recipe.views import autoscale
from starthinker_ui.recipe.models import Recipe, utc_milliseconds, JOB_LOOKBACK_MS


class Command(BaseCommand):
  help = 'Autoscale workers.'

  def handle(self, *args, **kwargs):
    print(json.dumps(json.loads(autoscale(None).content), indent=2))

    for recipe in Recipe.objects.filter(
        active=True, job_utm__lt=utc_milliseconds()).exclude(job_utm=0):
      print(recipe.id, recipe.name, recipe.get_days())
      print('---')

    for recipe in Recipe.objects.filter(
        active=True, worker_utm__gte=utc_milliseconds() - JOB_LOOKBACK_MS):
      print(recipe.id, recipe.name, recipe.worker_uid)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from django import forms
from django.db.models import Q

from starthinker_ui.recipe.models import Recipe
from starthinker_ui.project.models import Project
from starthinker_ui.recipe.forms_fields import ListChoiceField, ListChoiceIntegerField, TimezoneField

DAYS = [
    'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday', 'Sunday'
]
HOURS = [
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20,
    21, 22, 23
]


# This form is always constructed with an object, so initial data doesn't really work here, need to set defaults at the model
class SetupForm(forms.ModelForm):
  week = ListChoiceField(
      choices=map(lambda d: (d[:3], d), DAYS),
      initial=map(lambda d: d[:3], DAYS))
  hour = ListChoiceIntegerField(
      choices=map(lambda h: (h, h), HOURS), initial=[3])
  timezone = TimezoneField(required=False)
  project = forms.ModelChoiceField(queryset=None, required=False)

  class Meta:
    model = Recipe
    fields = ['name', 'project', 'timezone', 'week', 'hour', 'active']

  def __init__(self, manual, account, *args, **kwargs):
    super(SetupForm, self).__init__(*args, **kwargs)
    self.instance.account = account
    self.fields['active'].required = False

    query = Q(account=account) | Q(share='global')
    if account.get_domain():
      query |= (
          Q(share='domain') & ~Q(account__domain='')
          & Q(account__domain=account.get_domain()))
    self.fields['project'].queryset = Project.objects.filter(query).order_by(
        'share', 'identifier')

    self.fields['name'].help_text = 'Identify this recipe in list of recipes.'
    self.fields[
        'project'].help_text = ('Choose a <b>Google Cloud Project Service '
                                'Credential</b> uploaded to Projects.')
    self.fields[
        'timezone'].help_text = 'Frame of reference for all recipe times.'
    self.fields['week'].help_text = 'Days of week to execute recipe.'
    self.fields['hour'].help_text = 'Hours of day to execute recipe.'
    self.fields['active'].help_text = 'To pause recipe, uncheck this.'

    self.structure = [{
        'title': '%s Recipe' % self.instance.name.title(),
        'description': '',
        'fields': [self['name'], self['project']]
    }]

    if manual:
      del self.fields['week']
      del self.fields['hour']
      del self.fields['active']

      self.instance.week = []
      self.instance.hour = []
      self.instance.active = True
      self.instance.manual = True

      self.structure[0]['fields'].append(self['timezone'])

    else:
      self.structure.append({
          'title':
              'Schedule',
          'description':
              '',
          'fields': [
              self['timezone'],
              self['week'],
              self['hour'],
              self['active'],
          ]
      })



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import time
import csv
import io
import zipfile
import io
from datetime import date

from starthinker.util.project import project
from googleapiclient.http import MediaIoBaseDownload
from starthinker.util.data import get_rows
from starthinker.util.csv import column_header_sanitize, csv_to_rows
from starthinker.util.bigquery import rows_to_table, table_create, table_exists
from starthinker.util.sdf.schema.Lookup import SDF_Field_Lookup
from starthinker.util.google_api import API_DV360


# Desired file name: InsertionOrders, LineItems, *Camel case of the filetype
def get_single_sdf_rows(auth, version, partner_id, file_types, filter_type,
                        filter_ids_obj, desired_file_type):
  sdf_zip_file = sdf_download(auth, version, partner_id, file_types,
                              filter_type, filter_ids_obj)

  with zipfile.ZipFile(sdf_zip_file, 'r', zipfile.ZIP_DEFLATED) as d:
    file_names = d.namelist()
    for file_name in file_names:

      # make sure to only get the one file
      if desired_file_type != file_name.split('-')[1].split(
          '.')[0] or 'Skipped' in file_name:
        continue

      if project.verbose:
        print('SDF: Loading: ' + file_name)
      with d.open(file_name) as sdf_file:
        rows = csv_to_rows(sdf_file.read().decode('utf-8'))

        return rows


def sdf_download(auth, version, partner_id, file_types, filter_type,
                 filter_ids_obj):
  #Read Filter Ids
  filter_ids = list(get_rows(auth, filter_ids_obj))

  body = {
      'version': version,
      'partnerId': partner_id,
      'parentEntityFilter': {
          'fileType': file_types,
          'filterType': filter_type,
          'filterIds': filter_ids
      },
      'idFilter': None
  }

  operation = API_DV360(auth).sdfdownloadtasks().create(body=body).execute()

  if operation and 'name' in operation:
    request = API_DV360(auth).sdfdownloadtasks().operations().get(
        name=operation['name'])

    # This is the eng recommended way of getting the operation
    while True:
      response = request.execute()
      if 'done' in response and response['done']:
        break
      time.sleep(30)
  else:
    print('error')

  if 'error' in response:
    raise Exception(response['error']['message'])

  return download_media('user', response['response']['resourceName'])


def add_seekable_to_file(f):
  if not hasattr(f, 'seekable'):
    f.seekable = lambda: True


def sdf_to_bigquery(sdf_zip_file,
                    project_id,
                    dataset,
                    time_partitioned_table,
                    create_single_day_table,
                    table_suffix=''):
  with zipfile.ZipFile(sdf_zip_file, 'r', zipfile.ZIP_DEFLATED) as d:
    file_names = d.namelist()
    for file_name in file_names:
      if project.verbose:
        print('SDF: Loading: ' + file_name)
      with d.open(file_name) as sdf_file:
        rows = csv_to_rows(sdf_file.read().decode('utf-8'))
        if not rows:
          if project.verbose:
            print('SDF: Empty file ' + file_name)
          continue
        table_name = file_name.split('.')[0].replace('-', '_') + table_suffix
        schema = sdf_schema(next(rows))

        # Check if each SDF should have a dated table
        if create_single_day_table:
          table_name_dated = table_name + date.today().strftime('%Y_%m_%d')

          # Create table and upload data
          table_create('service', project_id, dataset, table_name_dated)
          rows_to_table(
              'service',
              project_id,
              dataset,
              table_name_dated,
              rows,
              schema=schema,
              skip_rows=1,
              disposition='WRITE_TRUNCATE')

        # Create end result table if it doesn't already exist
        if not table_exists('service', project_id, dataset, table_name):
          table_create(
              'service',
              project_id,
              dataset,
              table_name,
              is_time_partition=time_partitioned_table)

        rows_to_table(
            'service',
            project_id,
            dataset,
            table_name,
            rows,
            schema=schema,
            skip_rows=1,
            disposition='WRITE_APPEND'
            if time_partitioned_table else 'WRITE_TRUNCATE')


def sdf_schema(header):
  schema = []

  for h in header:
    schema.append({
        'name': column_header_sanitize(h),
        'type': SDF_Field_Lookup.get(h, 'STRING'),
        'mode': 'NULLABLE'
    })

  return schema


def download_media(auth, resource_name):
  if project.verbose:
    print('SDF: Start Download')

  downloadRequest = API_DV360(auth).media().download_media(
      resourceName=resource_name).execute(run=False)

  # Create output stream for downloaded file
  outStream = io.BytesIO()

  # Make downloader object
  downloader = MediaIoBaseDownload(outStream, downloadRequest)

  # Download media file in chunks until finished
  download_finished = False
  while download_finished is False:
    _, download_finished = downloader.next_chunk()

  if project.verbose:
    print('SDF: End Download')

  return outStream



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.bigquery import query_to_view
from starthinker.util.bigquery import table_create
from starthinker.util.data import get_rows
from starthinker.util.data import put_rows
from starthinker.util.google_api.discovery_to_bigquery import Discovery_To_BigQuery
from starthinker.util.project import project
from starthinker.util.regexp import lookup_id
from starthinker.util.sheets import sheets_clear

from starthinker.task.dv_editor.insertion_order import insertion_order_commit
from starthinker.task.dv_editor.line_item import line_item_commit
from starthinker.task.dv_editor.patch import patch_masks
from starthinker.task.dv_editor.patch import patch_preview


def partner_cost_clear():
  sheets_clear(project.task["auth_sheets"], project.task["sheet"], "Partner Costs",
               "A2:Z")


def partner_cost_load():

  # write partner_costs to sheet
  rows = get_rows(
      project.task["auth_bigquery"], {
          "bigquery": {
              "dataset":
                  project.task["dataset"],
              "query":
                  """SELECT
         CONCAT(P.displayName, ' - ', P.partnerId),
         CONCAT(A.displayName, ' - ', A.advertiserId),
         CONCAT(C.displayName, ' - ', C.campaignId),
         CONCAT(I.displayName, ' - ', I.insertionOrderId),
         NULL,
         PC.costType,
         PC.costType,
         PC.feeType,
         PC.feeType,
         PC.invoiceType,
         PC.invoiceType,
         PC.feeAmount / 100000,
         PC.feeAmount / 100000,
         PC.feePercentageMillis / 1000,
         PC.feePercentageMillis / 1000
       FROM `{dataset}.DV_InsertionOrders` AS I, UNNEST(partnerCosts) AS PC
       LEFT JOIN `{dataset}.DV_Campaigns` AS C
       ON I.campaignId=C.campaignId
       LEFT JOIN `{dataset}.DV_Advertisers` AS A
       ON I.advertiserId=A.advertiserId
       LEFT JOIN `{dataset}.DV_Partners` AS P
       ON A.partnerId=P.partnerId
       UNION ALL
       SELECT
         CONCAT(P.displayName, ' - ', P.partnerId),
         CONCAT(A.displayName, ' - ', A.advertiserId),
         CONCAT(C.displayName, ' - ', C.campaignId),
         CONCAT(I.displayName, ' - ', I.insertionOrderId),
         CONCAT(L.displayName, ' - ', L.lineItemId),
         PC.costType,
         PC.costType,
         PC.feeType,
         PC.feeType,
         PC.invoiceType,
         PC.invoiceType,
         PC.feeAmount / 100000,
         PC.feeAmount / 100000,
         PC.feePercentageMillis / 1000,
         PC.feePercentageMillis / 1000
       FROM `{dataset}.DV_LineItems` AS L, UNNEST(partnerCosts) AS PC
       LEFT JOIN `{dataset}.DV_Campaigns` AS C
       ON L.campaignId=C.campaignId
       LEFT JOIN `{dataset}.DV_InsertionOrders` AS I
       ON L.insertionOrderId=I.insertionOrderId
       LEFT JOIN `{dataset}.DV_Advertisers` AS A
       ON L.advertiserId=A.advertiserId
       LEFT JOIN `{dataset}.DV_Partners` AS P
       ON A.partnerId=P.partnerId
       """.format(**project.task),
              "legacy":
                  False
          }
      })

  put_rows(
      project.task["auth_sheets"], {
          "sheets": {
              "sheet": project.task["sheet"],
              "tab": "Partner Costs",
              "range": "A2"
          }
      }, rows)


def partner_cost_audit():
  rows = get_rows(
      project.task["auth_sheets"], {
          "sheets": {
              "sheet": project.task["sheet"],
              "tab": "Partner Costs",
              "range": "A2:Z"
          }
      })

  put_rows(
      project.task["auth_bigquery"], {
          "bigquery": {
              "dataset": project.task["dataset"],
              "table": "SHEET_PartnerCosts",
              "schema": [
                  { "name": "Partner", "type": "STRING" },
                  { "name": "Advertiser", "type": "STRING" },
                  { "name": "Campaign", "type": "STRING" },
                  { "name": "Insertion_Order", "type": "STRING" },
                  { "name": "Line_Item", "type": "STRING" },
                  { "name": "Cost_Type", "type": "STRING" },
                  { "name": "Cost_Type_Edit", "type": "STRING" },
                  { "name": "Fee_Type", "type": "STRING" },
                  { "name": "Fee_Type_Edit", "type": "STRING" },
                  { "name": "Invoice_Type", "type": "STRING" },
                  { "name": "Invoice_Type_Edit", "type": "STRING" },
                  { "name": "Fee_Amount", "type": "FLOAT" },
                  { "name": "Fee_Amount_Edit", "type": "FLOAT" },
                  { "name": "Fee_Percent", "type": "FLOAT" },
                  { "name": "Fee_Percent_Edit", "type": "FLOAT" },
              ],
              "format": "CSV"
          }
      }, rows)

  query_to_view(
      project.task["auth_bigquery"],
      project.id,
      project.task["dataset"],
      "AUDIT_PartnerCosts",
      """WITH
      /* Check if sheet values are set */
      INPUT_ERRORS AS (
        SELECT
        *
        FROM (
          SELECT
            'Partner Costs' AS Operation,
            CASE
              WHEN Cost_Type_Edit IS NULL THEN 'Missing Cost Type.'
              WHEN Fee_Type_Edit IS NULL THEN 'Missing Fee Type.'
              WHEN Invoice_Type_Edit IS NULL THEN 'Missing Invoice Type.'
              WHEN Fee_Amount_Edit IS NULL THEN 'Missing Fee Amount.'
              WHEN Fee_Percent_Edit IS NULL THEN 'Missing Fee Percent.'
            ELSE
              NULL
            END AS Error,
            'ERROR' AS Severity,
          COALESCE(Line_Item, Insertion_Order, 'BLANK') AS Id
        FROM
          `{dataset}.SHEET_PartnerCosts`
        )
        WHERE
          Error IS NOT NULL
      )

      SELECT * FROM INPUT_ERRORS
      ;
    """.format(**project.task),
      legacy=False)

  query_to_view(
    project.task["auth_bigquery"],
    project.id,
    project.task["dataset"],
    "PATCH_PartnerCosts",
    """SELECT *
      FROM `{dataset}.SHEET_PartnerCosts`
      WHERE (
        REGEXP_CONTAINS(Insertion_Order, r" - (\d+)$")
        OR REGEXP_CONTAINS(Line_Item, r" - (\d+)$")
      )
      AND Line_Item NOT IN (SELECT Id FROM `{dataset}.AUDIT_PartnerCosts` WHERE Severity='ERROR')
      AND Insertion_Order NOT IN (SELECT Id FROM `{dataset}.AUDIT_PartnerCosts` WHERE Severity='ERROR')
    """.format(**project.task),
    legacy=False
  )


def partner_cost_patch(commit=False):
  patches = {}
  changed = set()

  rows = get_rows(
    project.task["auth_bigquery"],
    { "bigquery": {
      "dataset": project.task["dataset"],
      "table":"PATCH_PartnerCosts",
    }},
    as_object=True
  )

  for row in rows:

    lookup = row['Line_Item'] or row['Insertion_Order']

    patches.setdefault(
        lookup, {
            "operation": "Partner Costs",
            "action": "PATCH",
            "partner": row['Partner'],
            "advertiser": row['Advertiser'],
            "campaign": row['Campaign'],
            "parameters": {
                "advertiserId": lookup_id(row['Advertiser']),
                "body": {
                    "partnerCosts": []
                }
            }
        })

    if row['Line_Item']:
      patches[lookup]["line_item"] = row['Line_Item']
      patches[lookup]["parameters"]["lineItemId"] = lookup_id(row['Line_Item'])
    else:
      patches[lookup]["insertion_order"] = row['Insertion_Order']
      patches[lookup]["parameters"]["insertionOrderId"] = lookup_id(row['Insertion_Order'])

    patches[lookup]["parameters"]["body"]["partnerCosts"].append({
        "costType": row['Cost_Type_Edit'],
        "feeType": row['Fee_Type_Edit'],
        "invoiceType": row['Invoice_Type_Edit'],
        "feeAmount": int(float(row['Fee_Amount_Edit']) * 100000) if row['Fee_Amount_Edit'] else None,
        "feePercentageMillis": int(float(row['Fee_Percent_Edit']) * 1000) if row['Fee_Percent_Edit'] else None
    })

    if row['Cost_Type'] != row['Cost_Type_Edit'] \
      or row['Fee_Type'] != row['Fee_Type_Edit'] \
      or row['Invoice_Type'] != row['Invoice_Type_Edit'] \
      or row['Fee_Amount'] != row['Fee_Amount_Edit'] \
      or row['Fee_Percent'] != row['Fee_Percent_Edit']:
      changed.add(lookup)

  # Remove any patches where partner costs have not changed
  for pc in list(patches.keys()):
    if pc not in changed:
      del patches[pc]
  patches = list(patches.values())

  patch_masks(patches)

  if commit:
    insertion_order_commit(patches)
    line_item_commit(patches)
  else:
    patch_preview(patches)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from django.apps import AppConfig
from django.conf import settings

from starthinker.util.project import project
from starthinker.util.storage import bucket_create

USER_BUCKET = '%s-starthinker-credentials' % settings.UI_PROJECT.split(
    ':', 1)[-1]  # remove domain: part
USER_LOCATION = settings.UI_ZONE.rsplit('-',
                                        1)[0]  # take only region part of zone


class AccountConfig(AppConfig):
  name = 'starthinker_ui.account'

  def ready(self):
    print('CHECKING IF USER BUCKET EXISTS:', USER_BUCKET, USER_LOCATION)
    project.initialize(
        _project=settings.UI_PROJECT, _service=settings.UI_SERVICE)
    bucket_create('service', settings.UI_PROJECT, USER_BUCKET, USER_LOCATION)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import base64

from starthinker.util.google_api import API_PubSub


def topic_create(suth, project_id, topic):
  api = API_PubSub(auth).projects().topics().create(
      topic='projects/%s/topics/%s' % (project_id, topic), body={}).execute()


def subscription_create(auth, project_id, topic, subscription):
  body = {
      'topic': 'projects/%s/topics/%s' % (project_id, topic),
      'pushConfig': {},
      'ackDeadlineSeconds': 600,
      'messageRetentionDuration': '86400s',  # 24 hours
      'retainAckedMessages': False,
  }

  api = API_PubSub(auth).projects().subscriptions().create(
      name='projects/%s/subscriptions/%s' % (project_id, subscription),
      body=body).execute()


def topic_publish(auth, project_id, topic, data):
  body = {'messages': [{'data': base64.b64encode(data)}]}
  api = API_PubSub(auth).projects().topics().publish(
      topic='projects/%s/topics/%s' % (project_id, topic), body=body).execute()
  return api['messageIds'][0]


def subscription_acknowledge(auth, project_id, subscription, ack_id):
  if isinstance(ack_id, str):
    ack_id = [ack_id] if ack_id else []

  if ack_id:
    body = {'ackIds': ack_id}
    API_PubSub(auth).projects().subscriptions().acknowledge(
        subscription='projects/%s/subscriptions/%s' %
        (project_id, subscription),
        body=body).execute()


def subscription_pull(auth,
                      project_id,
                      subscription,
                      immediate=True,
                      maximum=1,
                      acknowledge=False):
  messages = []

  if maximum <= 0:
    return messages

  body = {'returnImmediately': immediate, 'maxMessages': maximum}

  for message in API_PubSub(
      auth, iterate=True).projects().subscriptions().pull(
          subscription='projects/%s/subscriptions/%s' %
          (project_id, subscription),
          body=body).execute():
    messages.append({
        'ackId': message['ackId'],
        'data': base64.b64decode(message['message']['data'])
    })

  # if acknowledge, then acknowledge all messages and return only data
  if acknowledge:
    subscription_acknowledge(auth, project_id, subscription,
                             [m['ackId'] for m in messages])
    return [m['data'] for m in messages]
  # or return ack and data as a dictionary
  else:
    return messages



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
""" Handler that executes { "iam":{...}} task in recipe JSON.

Grants roles to users in Google Cloud Projects.  Typically you define a role
in the cloud project that grants a collection of permissions for assets in
that project.  Use this call to quickly grant a user a role.

Users who authenticate against your client credentials DO NOT automatically get
access permission.  That still needs to be granted on a case by case basis.
Placing a call to this handler early in a recipe JSON ensures the user executing
the recipe has the right privileges.

This handler should be called with "service" auth in the the JSON.  The service
should
be able to assign roles.

### Command Line COnvenience

Mostly a helper function, your service credential will already have a higher
level
of role granting and you need it to grant the role to your user.  So there is
NO SECURITY benefit, just a conveniece.

### UI Security

In a UI environment such as a web application, where users DO NOT have access to
the service credentials, but the server does have access to user credentials,
this
handler allows the service to securely grant additional roles to users.

### Good Practice

Using roles is a better practice than assigning permissions to user accounts
individually because it allows for better tracking and quicker revocation.
Changing a role permission changes all user at once without having to track
down individuals.  We highly recommend using roles.

"""

from starthinker.util.project import project
from starthinker.util.auth import set_iam


@project.from_parameters
def iam():
  set_iam(project.task['auth'], project.id, project.task['role'],
          project.task['email'])


if __name__ == '__main__':
  iam()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import json

from django.db import models

from starthinker_ui.account.models import Account


class Project(models.Model):
  account = models.ForeignKey(Account, on_delete=models.PROTECT, null=True)
  identifier = models.CharField(max_length=255)
  service = models.TextField(default='')
  key = models.CharField(max_length=50, default='')
  share = models.CharField(max_length=50, default='')

  def __str__(self):
    if self.share == 'domain':
      return 'DOMAIN: %s | CAUTION: Be sure you trust this project.' % self.identifier
    elif self.share == 'global':
      return 'GLOBAL: %s | CAUTION: Be sure you trust this project.' % self.identifier
    else:
      return 'USER: %s | SAFE: Only you are using this service account.' % self.identifier

  def link_edit(self):
    return '/project/edit/%d/' % self.pk

  def link_delete(self):
    return '/project/delete/%d/' % self.pk

  def get_credentials_service(self):
    return self.service if self.service else '{}'

  def get_client_email(self):
    return json.loads(self.get_credentials_service()).get('client_email', '')

  def get_project_id(self):
    return json.loads(self.get_credentials_service()).get('project_id', '')



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

Floodlight_Report_Metrics_Schema = [{
    'name': 'Click_Through_Conversions',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Click_Through_Revenue',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'View_Through_Conversions',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'View_Through_Revenue',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_1',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_2',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_3',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_4',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_5',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_6',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_7',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_8',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_9',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_10',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_11',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_12',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_13',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_14',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_15',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_16',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_17',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_18',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_19',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_20',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_21',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_22',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_23',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_24',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_25',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_26',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_27',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_28',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_29',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_30',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_31',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_32',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_33',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_34',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_35',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_36',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_37',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_38',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_39',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_40',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_41',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_42',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_43',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_44',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_45',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_46',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_47',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_48',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_49',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_50',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_51',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_52',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_53',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_54',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_55',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_56',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_57',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_58',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_59',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_60',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_61',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_62',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_63',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_64',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_65',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_66',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_67',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_68',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_69',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_70',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_71',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_72',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_73',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_74',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_75',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_76',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_77',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_78',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_79',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_80',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_81',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_82',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_83',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_84',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_85',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_86',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_87',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_88',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_89',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_90',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_91',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_92',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_93',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_94',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_95',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_96',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_97',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_98',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_99',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_100',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Click_Through_Transaction_Count',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'View_Through_Transaction_Count',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Average_Expansion_Time',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Full_Screen_Average_View_Time',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Average_Interaction_Time',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Average_View_Time',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Backup_Image',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Average_Time',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Counters',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Exits',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Timers',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Total_Display_Time',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Expansion_Time',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Expansions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Full_Screen_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Full_Screen_Video_Completions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Full_Screen_Video_Plays',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Html5_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Total_Interaction_Time',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Total_Interactions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Interactive_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Manual_Closes',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Companion_Clicks',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Completions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_First_Quartile_Completions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Full_Screen',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Interaction_Rate',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Interactions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Midpoints',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Mutes',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Pauses',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Plays',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Progress_Events',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Replays',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Skips',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Stops',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Third_Quartile_Completions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Unmutes',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_View_Rate',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Views',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Total_Conversions',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Total_Revenue',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Transaction_Count',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}]



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Handles creation and updates of dynamic targeting keys."""

from starthinker.task.traffic.dao import BaseDAO
from starthinker.task.traffic.feed import FieldMap


class DynamicTargetingKeyDAO(BaseDAO):
  """Landing page data access object.

  Inherits from BaseDAO and implements dynamic targeting key specific logic for
  creating
  and
  updating dynamic targeting key.
  """

  def __init__(self, auth, profile_id, is_admin):
    """Initializes DynamicTargetingKeyDAO with profile id and authentication scheme."""

    super(DynamicTargetingKeyDAO, self).__init__(auth, profile_id, is_admin)

    self._id_field = None
    self._search_field = None
    self._list_name = 'dynamicTargetingKeys'
    self._entity = 'DYNAMIC_TARGETING_KEY'
    self._parent_filter_name = None
    self._parent_filter_field_name = None
    self._parent_dao = None
    self._key_cache = {}

  def _api(self, iterate=False):
    """Returns an DCM API instance for this DAO."""
    return super(DynamicTargetingKeyDAO,
                 self)._api(iterate).dynamicTargetingKeys()

  def _key_exists(self, advertiser_id, key_name):
    cache_key = str(advertiser_id) + key_name

    if not cache_key in self._key_cache:
      keys = self._api().list(
          profileId=self.profile_id,
          advertiserId=advertiser_id,
          names=[key_name]).execute()

      self._key_cache[cache_key] = 'dynamicTargetingKeys' in keys and len(
          keys['dynamicTargetingKeys']) > 0

    return self._key_cache[cache_key]

  def _create_key(self, key_name, object_type, object_id):
    key = {'name': key_name, 'objectId': object_id, 'objectType': object_type}

    self._api().insert(profileId=self.profile_id, body=key).execute()

    if object_type == 'OBJECT_ADVERTISER':
      cache_key = str(object_id) + key_name
      self._key_cache[cache_key] = True

  def process(self, feed_item):
    """Processes a Bulkdozer feed item.

    This method identifies if the dyanmic targeting key already exists in CM, if
    it doesn't it creates it associated with the advertiser, and then inserts an
    association with the identified object.

    Args:
      feed_item: Bulkdozer feed item to process.

    Returns:
      Newly created or updated CM object.
    """
    if feed_item.get(FieldMap.ADVERTISER_ID, None) and feed_item.get(
        FieldMap.DYNAMIC_TARGETING_KEY_NAME, None):
      if not self._key_exists(
          feed_item.get(FieldMap.ADVERTISER_ID, None),
          feed_item.get(FieldMap.DYNAMIC_TARGETING_KEY_NAME, None)):
        self._create_key(
            feed_item.get(FieldMap.DYNAMIC_TARGETING_KEY_NAME, None),
            'OBJECT_ADVERTISER', feed_item.get(FieldMap.ADVERTISER_ID, None))

      self._create_key(
          feed_item.get(FieldMap.DYNAMIC_TARGETING_KEY_NAME, None),
          feed_item.get(FieldMap.DYNAMIC_TARGETING_KEY_OBJECT_TYPE, None),
          feed_item.get(FieldMap.DYNAMIC_TARGETING_KEY_OBJECT_ID, None))
    else:
      raise Exception(
          'Dynamic targeting key, %s and %s are required' %
          (FieldMap.ADVERTISER_ID, FieldMap.DYNAMIC_TARGETING_KEY_NAME))



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import re
import zlib
import gzip
from io import BytesIO
from datetime import datetime, timedelta
from googleapiclient.errors import HttpError

from starthinker.util.project import project
from starthinker.util.bigquery import io_to_table, storage_to_table, table_exists, job_wait
from starthinker.util.storage import object_list, object_get_chunks
from starthinker.util.csv import column_header_sanitize
from starthinker.task.dt.schema.Lookup import DT_Field_Lookup

HEADER_SIZE = 1024 * 10  # 10K should be enough for longest possible DT header
RE_DT_TIME = re.compile(r'.+?_(\d+)_\d+_\d+_\d+\.csv\.gz')


def dt_schema(header):
  schema = []
  for h in header:
    h = column_header_sanitize(h)
    schema.append({
        'name': h,
        'type': DT_Field_Lookup.get(h, 'STRING'),
        'mode': 'NULLABLE'
    })
  return schema


def dt_header(dt_file):
  if project.verbose:
    print('DT HEADER')

  # find first dt file to match pattern
  path = '%s:%s' % (project.task['bucket'], dt_file)

  # find first line of file ( gzip will decompress partial, and pull header out )
  sample_data = next(object_get_chunks(project.task['auth'], path, HEADER_SIZE))
  with gzip.GzipFile(fileobj=BytesIO(sample_data), mode='rb') as fo:
    sample_header = fo.read(HEADER_SIZE).decode('utf-8').split('\n')[0]

  return sample_header.split(',')


def dt_timestamp(dt_filepath):
  dt_time = None

  parts = RE_DT_TIME.match(dt_filepath)

  if parts:
    dt_time = parts.groups()[0]
    if len(dt_time) == 10:
      dt_time = project.timezone.localize(
          datetime.strptime(dt_time, '%Y%m%d%H'))
    else:
      dt_time = project.timezone.localize(datetime.strptime(dt_time, '%Y%m%d'))

  return dt_time


def dt_move_large(dt_file, dt_partition, jobs):
  if project.verbose:
    print('DT TO TABLE LARGE', dt_partition)

  delimiter = '\n'
  disposition = 'WRITE_TRUNCATE'

  # decompression handler for gzip ( must be outside of chunks as it keeps track of stream across multiple calls )
  gz_handler = zlib.decompressobj(32 + zlib.MAX_WBITS)

  # sliding view of data flowing out of decompression, used to buffer and delimit rows
  first_row = True
  view = ''

  # loop all chunks of file, decompress, and find row delimiter
  for data_gz in object_get_chunks(project.task['auth'],
                                   '%s:%s' % (project.task['bucket'], dt_file)):

    view += gz_handler.decompress(data_gz.read()).decode('utf-8')

    if first_row:
      end = view.find(delimiter)
      schema = dt_schema(view[:end].split(','))
      view = view[(end + 1):]
      first_row = False

    end = view.rfind(delimiter)

    jobs.append(
        io_to_table(project.task['auth'], project.id,
                    project.task['to']['dataset'], dt_partition,
                    BytesIO(view[:end].encode()), 'CSV', schema, 0, disposition,
                    False))
    disposition = 'WRITE_APPEND'
    view = view[min(end + 1, len(view)):]


def dt_move_small(dt_file, dt_partition, jobs):
  if project.verbose:
    print('DT TO TABLE SMALL', dt_partition)

  jobs.append(
      storage_to_table(project.task['auth'], project.id,
                       project.task['to']['dataset'], dt_partition,
                       '%s:%s*' % (project.task['bucket'], dt_file),
                       dt_schema(dt_header(dt_file)), 1, 'CSV',
                       'WRITE_TRUNCATE', False))


def dt_move(dt_object, dt_partition, jobs):
  """ Due to BQ limit, files over 5 GB need to be transfered using inline decompression ( slower but works 100% of the time )"""
  if int(dt_object['size']) > 5000000:
    dt_move_large(dt_object['name'], dt_partition, jobs)
  else:
    dt_move_small(dt_object['name'], dt_partition, jobs)


@project.from_parameters
def dt():
  jobs = []

  if project.verbose:
    print('DT To BigQuery')

  # legacy deprecated ( do not use )
  if 'path' in project.task:
    project.task['paths'] = [project.task['path']]

  # loop all dt files to match pattern or match any pattern
  print('PATHS', project.task['paths'])

  for path in (project.task['paths'] or ['']):

    print(path)
    for dt_object in object_list(
        project.task['auth'],
        '%s:%s' % (project.task['bucket'], path),
        raw=True):
      dt_size = dt_object['size']
      dt_file = dt_object['name']
      dt_time = dt_timestamp(dt_file)

      dt_partition = dt_file.split('.', 1)[0]
      if ((project.task.get('days') is None and
           project.task.get('hours') is None) or
          (dt_time > project.now - timedelta(
              days=project.task.get('days', 60),
              hours=project.task.get('hours', 0)))):
        if not table_exists(project.task['to']['auth'], project.id,
                            project.task['to']['dataset'], dt_partition):
          dt_move(dt_object, dt_partition, jobs)
        else:
          if project.verbose:
            print('DT Partition Exists:', dt_partition)

  for count, job in enumerate(jobs):
    print('Waiting For Job: %d of %d' % (count + 1, len(jobs)))
    job_wait(project.task['to']['auth'], job)


if __name__ == '__main__':
  dt()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer ( recommended ):

    From Release: pip install starthinker
    From Open Source: pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory ( if pushing local code changes ):

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

  If any recipe task has "auth" set to "user" add user credentials:

    1. Ensure an RECIPE['setup']['auth']['user'] = [User Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_user", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/deploy_commandline.md#optional-setup-user-credentials

--------------------------------------------------------------

  If any recipe task has "auth" set to "service" add service credentials:

    1. Ensure an RECIPE['setup']['auth']['service'] = [Service Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_service", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md

--------------------------------------------------------------

Bulkdozer

Bulkdozer is a tool that can reduce trafficking time in Campaign Manager by up to 80%% by providing automated bulk editing capabilities.

  - Open the <a href='https://docs.google.com/spreadsheets/d/10YewffRUP1gCxTY0ZLTkVEfzmOQ1gPXPxp7qev3w8xk/edit?usp=sharing' target='_blank'>Bulkdozer 0.27</a> feed.
  - Make your own copy of the feed by clicking the File -> Make a copy... menu in the feed.
  - Give it a meaninful name including the version, your name, and team to help you identify it and ensure you are using the correct version.
  - Under the Account ID field below, enter the your Campaign Manager Network ID.
  - Under Sheet URL, enter the URL of your copy of the feed that you just created in the steps above.
  - Go to the Store tab of your new feed, and enter your profile ID in the profileId field (cell B2). Your profile ID is visible in Campaign Manager by clicking your avatar on the top right corner.
  - Click the 'Save' button below.
  - After clicking 'Save', copy this page's URL from your browser address bar, and paste it in the Store tab for the recipe_url field (cell B5) your sheet.
  - Bulkdozer is ready for use
  - Stay up to date on new releases and other general anouncements by joining <a href='https://groups.google.com/forum/#!forum/bulkdozer-announcements' target='_blank'>Bulkdozer announcements</a>.
  - Review the <a href='https://github.com/google/starthinker/blob/master/tutorials/Bulkdozer/Installation_and_User_guides.md' target='_blank'>Bulkdozer documentation</a>.

--------------------------------------------------------------

This StarThinker DAG can be extended with any additional tasks from the following sources:
  - https://google.github.io/starthinker/
  - https://github.com/google/starthinker/tree/master/dags

'''

from starthinker.airflow.factory import DAG_Factory

INPUTS = {
  'recipe_timezone': 'America/Chicago',  # Timezone for report dates.
  'account_id': '',  # Campaign Manager Network ID (optional if profile id provided)
  'dcm_profile_id': '',  # Campaign Manager Profile ID (optional if account id provided)
  'sheet_url': '',  # Feed Sheet URL
}

RECIPE = {
  'setup': {
    'day': [
    ],
    'hour': [
    ]
  },
  'tasks': [
    {
      'traffic': {
        'hour': [
        ],
        'account_id': {
          'field': {
            'name': 'account_id',
            'kind': 'string',
            'order': 1,
            'description': 'Campaign Manager Network ID (optional if profile id provided)',
            'default': ''
          }
        },
        'dcm_profile_id': {
          'field': {
            'name': 'dcm_profile_id',
            'kind': 'string',
            'order': 1,
            'description': 'Campaign Manager Profile ID (optional if account id provided)',
            'default': ''
          }
        },
        'auth': 'user',
        'sheet_url': {
          'field': {
            'name': 'sheet_url',
            'kind': 'string',
            'order': 2,
            'description': 'Feed Sheet URL',
            'default': ''
          }
        },
        'timezone': {
          'field': {
            'name': 'recipe_timezone',
            'kind': 'timezone',
            'description': 'Timezone for report dates.',
            'default': 'America/Chicago'
          }
        }
      }
    }
  ]
}

DAG_FACTORY = DAG_Factory('bulkdozer', RECIPE, INPUTS)
DAG = DAG_FACTORY.generate()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer ( recommended ):

    From Release: pip install starthinker
    From Open Source: pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory ( if pushing local code changes ):

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

  If any recipe task has "auth" set to "user" add user credentials:

    1. Ensure an RECIPE['setup']['auth']['user'] = [User Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_user", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/deploy_commandline.md#optional-setup-user-credentials

--------------------------------------------------------------

  If any recipe task has "auth" set to "service" add service credentials:

    1. Ensure an RECIPE['setup']['auth']['service'] = [Service Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_service", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md

--------------------------------------------------------------

Google API To BigQuery

Execute any Google API function and store results to BigQuery.

  - Enter an api name and version.
  - Specify the function using dot notation.
  - Specify the arguments using json.
  - Iterate is optional, use if API returns a list of items that are not unpacking correctly.
  - The <a href='https://cloud.google.com/docs/authentication/api-keys' target='_blank'>API Key</a> may be required for some calls.
  - The <a href='https://developers.google.com/google-ads/api/docs/first-call/dev-token' target='_blank'>Developer Token</a> may be required for some calls.
  - Give BigQuery dataset and table where response will be written.
  - All API calls are based on <a href='https://developers.google.com/discovery/v1/reference'  target='_blank'>discovery document</a>, for example the <a href='https://developers.google.com/display-video/api/reference/rest/v1/advertisers/list' target='_blank'>Campaign Manager API</a>.

--------------------------------------------------------------

This StarThinker DAG can be extended with any additional tasks from the following sources:
  - https://google.github.io/starthinker/
  - https://github.com/google/starthinker/tree/master/dags

'''

from starthinker.airflow.factory import DAG_Factory

INPUTS = {
  'auth_read': 'user',  # Credentials used for reading data.
  'api': 'displayvideo',  # See developer guide.
  'version': 'v1',  # Must be supported version.
  'function': 'advertisers.list',  # Full function dot notation path.
  'kwargs': {'partnerId': 234340},  # Dictionray object of name value pairs.
  'api_key': '',  # Associated with a Google Cloud Project.
  'developer_token': '',  # Associated with your organization.
  'login_customer_id': '',  # Associated with your Adwords account.
  'dataset': '',  # Existing dataset in BigQuery.
  'table': '',  # Table to write API call results to.
}

RECIPE = {
  'tasks': [
    {
      'google_api': {
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 1,
            'default': 'user',
            'description': 'Credentials used for reading data.'
          }
        },
        'api': {
          'field': {
            'name': 'api',
            'kind': 'string',
            'order': 1,
            'default': 'displayvideo',
            'description': 'See developer guide.'
          }
        },
        'version': {
          'field': {
            'name': 'version',
            'kind': 'string',
            'order': 2,
            'default': 'v1',
            'description': 'Must be supported version.'
          }
        },
        'function': {
          'field': {
            'name': 'function',
            'kind': 'string',
            'order': 3,
            'default': 'advertisers.list',
            'description': 'Full function dot notation path.'
          }
        },
        'kwargs': {
          'field': {
            'name': 'kwargs',
            'kind': 'json',
            'order': 4,
            'default': {
              'partnerId': 234340
            },
            'description': 'Dictionray object of name value pairs.'
          }
        },
        'key': {
          'field': {
            'name': 'api_key',
            'kind': 'string',
            'order': 4,
            'default': '',
            'description': 'Associated with a Google Cloud Project.'
          }
        },
        'headers': {
          'developer-token': {
            'field': {
              'name': 'developer_token',
              'kind': 'string',
              'order': 5,
              'default': '',
              'description': 'Associated with your organization.'
            }
          },
          'login-customer-id': {
            'field': {
              'name': 'login_customer_id',
              'kind': 'string',
              'order': 6,
              'default': '',
              'description': 'Associated with your Adwords account.'
            }
          }
        },
        'results': {
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'dataset',
                'kind': 'string',
                'order': 7,
                'default': '',
                'description': 'Existing dataset in BigQuery.'
              }
            },
            'table': {
              'field': {
                'name': 'table',
                'kind': 'string',
                'order': 8,
                'default': '',
                'description': 'Table to write API call results to.'
              }
            }
          }
        }
      }
    }
  ]
}

DAG_FACTORY = DAG_Factory('google_api_to_bigquery', RECIPE, INPUTS)
DAG = DAG_FACTORY.generate()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



#{
#  "notify": {
#    "auth":"user",
#    "to": "mauriciod@google.com",
#    "query": "select pv.advertiser, pv.io, pv.io_id, string(pv.pacing * 100) + ' %'  from [starthinker.pacing_view] pv where pacing < 0.7 order by pv.pacing asc"
#  }
#}

from starthinker.util.project import project
from starthinker.util.email import send_email
from starthinker.util.bigquery import query_to_rows


@project.from_parameters
def notify():
  if project.verbose:
    print('Notify')

  email_to = project.task['to']
  auth = project.task['auth']
  dataset = project.task['dataset']
  sql = project.task['query']
  rows = [row for row in query_to_rows(auth, project.id, dataset, sql)]

  if rows:
    html = '<body><table border=1><tr>'

    for row in rows:
      html += '<tr>'

      for column in row:
        html += '<td>%s</td>' % column

      html += '</tr>'

    html += '</table></body>'
    send_email(auth, email_to, 'mauriciod@google.com', None,
               'You have underpacing campaigns!', None, html)


if __name__ == '__main__':
  notify()
